import os
import warnings

import numpy as np

from astropy.time import Time
import astropy.units as u
from functools import cached_property
from glue.core.message import EditSubsetMessage, SubsetUpdateMessage
from glue.core.edit_subset_mode import (AndMode, AndNotMode, OrMode,
                                        ReplaceMode, XorMode, NewMode)
from glue.core.roi import CircularROI, CircularAnnulusROI, EllipticalROI, RectangularROI
from glue.core.subset import (RoiSubsetState, RangeSubsetState, CompositeSubsetState,
                              MaskSubsetState)
from glue.icons import icon_path
from glue_jupyter.widgets.subset_mode_vuetify import SelectionModeMenu
from glue_jupyter.common.toolbar_vuetify import read_icon
from traitlets import Any, List, Unicode, Bool, observe

from specutils import SpectralRegion
from photutils.aperture import (CircularAperture, SkyCircularAperture,
                                EllipticalAperture, SkyEllipticalAperture,
                                RectangularAperture, SkyRectangularAperture,
                                CircularAnnulus, SkyCircularAnnulus)
from regions import (Regions, CirclePixelRegion, CircleSkyRegion,
                     EllipsePixelRegion, EllipseSkyRegion,
                     RectanglePixelRegion, RectangleSkyRegion,
                     CircleAnnulusPixelRegion, CircleAnnulusSkyRegion)

from jdaviz.core.region_translators import regions2roi, aperture2regions
from jdaviz.core.events import (SnackbarMessage, GlobalDisplayUnitChanged,
                                LinkUpdatedMessage, SubsetRenameMessage)
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, DatasetSelect,
                                        SubsetSelect, SelectPluginComponent,
                                        LoadersMixin)
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.helpers import _next_subset_num
from jdaviz.utils import MultiMaskSubsetState, _chain_regions, data_has_valid_wcs

from jdaviz.configs.default.plugins.subset_tools import utils


__all__ = ['SubsetTools']

SUBSET_MODES = {
    'new': NewMode,
    'replace': ReplaceMode,
    'OrState': OrMode,
    'AndState': AndMode,
    'XorState': XorMode,
    'AndNotState': AndNotMode,
    'RangeSubsetState': RangeSubsetState,
    'RoiSubsetState': RoiSubsetState
}
SUBSET_MODES_PRETTY = {
    'new': NewMode,
    'replace': ReplaceMode,
    'or': OrMode,
    'and': AndMode,
    'xor': XorMode,
    'andnot': AndNotMode,
}
SUBSET_TO_PRETTY = {v: k for k, v in SUBSET_MODES_PRETTY.items()}
COMBO_OPTIONS = list(SUBSET_MODES_PRETTY.keys())


@tray_registry('g-subset-tools', label="Subset Tools")
class SubsetTools(PluginTemplateMixin, LoadersMixin):
    """
    See the :ref:`Subset Tools <imviz-subset-plugin>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``subset`` (:class:`~jdaviz.core.template_mixin.SubsetSelect`):
      Manages subset selection and creation
    * ``combination_mode`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Allows selection of combination modes for subsets
    * ``recenter_dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Data used for recentering.
    * :meth:`recenter`
    * :meth:`get_center`
    * :meth:`set_center`
    * :meth:`import_region`
    * :meth:`get_regions`
    * :meth:`rename_selected`
    * :meth:`rename_subset`
    """
    template_file = __file__, "subset_tools.vue"
    select = List([]).tag(sync=True)
    subset_items = List([]).tag(sync=True)
    subset_selected = Any().tag(sync=True)

    mode_selected = Unicode('add').tag(sync=True)
    show_region_info = Bool(True).tag(sync=True)
    subset_types = List([]).tag(sync=True)
    subset_definitions = List([]).tag(sync=True)
    glue_state_types = List([]).tag(sync=True)
    has_subset_details = Bool(False).tag(sync=True)

    recenter_dataset_items = List().tag(sync=True)
    recenter_dataset_selected = Unicode().tag(sync=True)

    subplugins_opened = Any().tag(sync=True)

    multiselect = Bool(False).tag(sync=True)  # multiselect only for subset
    is_centerable = Bool(False).tag(sync=True)
    can_simplify = Bool(False).tag(sync=True)
    can_freeze = Bool(False).tag(sync=True)

    icon_replace = Unicode(read_icon(os.path.join(icon_path("glue_replace", icon_format="svg")), 'svg+xml')).tag(sync=True)  # noqa
    icon_or = Unicode(read_icon(os.path.join(icon_path("glue_or", icon_format="svg")), 'svg+xml')).tag(sync=True)  # noqa
    icon_and = Unicode(read_icon(os.path.join(icon_path("glue_and", icon_format="svg")), 'svg+xml')).tag(sync=True)  # noqa
    icon_xor = Unicode(read_icon(os.path.join(icon_path("glue_xor", icon_format="svg")), 'svg+xml')).tag(sync=True)  # noqa
    icon_andnot = Unicode(read_icon(os.path.join(icon_path("glue_andnot", icon_format="svg")), 'svg+xml')).tag(sync=True)  # noqa

    icon_radialtocheck = Unicode(read_icon(os.path.join(ICON_DIR, 'radialtocheck.svg'), 'svg+xml')).tag(sync=True)  # noqa
    icon_checktoradial = Unicode(read_icon(os.path.join(ICON_DIR, 'checktoradial.svg'), 'svg+xml')).tag(sync=True)  # noqa

    combination_mode_items = List([]).tag(sync=True)
    combination_mode_selected = Any().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        config = self.app.config
        if config in ['cubeviz', 'specviz2d', 'mosviz', 'rampviz']:
            self._plugin_description = 'Select and interact with spectral/spatial subsets.'
        elif config == 'imviz':
            self._plugin_description = 'Select and interact with spatial subsets.'
        elif config == 'specviz':
            self._plugin_description = 'Select and interact with spectral subsets.'

        self.components = {
            'g-subset-mode': SelectionModeMenu(session=self.session)
        }

        self.session.hub.subscribe(self, EditSubsetMessage,
                                   handler=self._sync_selected_from_state)
        self.session.hub.subscribe(self, SubsetUpdateMessage,
                                   handler=self._on_subset_update)
        self.session.hub.subscribe(self, GlobalDisplayUnitChanged,
                                   handler=self._on_display_unit_changed)
        self.session.hub.subscribe(self, LinkUpdatedMessage,
                                   handler=self._on_link_update)
        self.session.hub.subscribe(self, SubsetRenameMessage,
                                   handler=self._sync_available_from_state)

        self.subset = SubsetSelect(self,
                                   'subset_items',
                                   'subset_selected',
                                   multiselect='multiselect',
                                   default_text="Create New")
        self.subset_states = []
        self.spectral_display_unit = None

        align_by = getattr(self.app, '_align_by', None)
        self.display_sky_coordinates = (align_by == 'wcs' and not self.multiselect)

        self.recenter_dataset = DatasetSelect(self, 'recenter_dataset_items',
                                              'recenter_dataset_selected',
                                              multiselect=None)

        self.combination_mode = SelectPluginComponent(self,
                                                      items='combination_mode_items',
                                                      selected='combination_mode_selected',
                                                      manual_options=COMBO_OPTIONS)

    @property
    def user_api(self):
        expose = ['subset', 'combination_mode',
                  'recenter_dataset', 'recenter',
                  'get_center', 'set_center',
                  'import_region', 'get_regions',
                  'rename_selected', 'rename_subset',
                  'update_subset']
        if self.dev_loaders:
            expose += ['loaders']
        return PluginUserApi(self, expose)

    def get_regions(self, region_type=None, list_of_subset_labels=None,
                    use_display_units=False):
        """
        Return spatial and/or spectral subsets of ``region_type`` (spatial or
        spectral, default both) as ``regions`` or ``SpectralRegions`` objects,
        respectivley.

        Parameters
        ----------
        region_type : str or None, optional
            Specifies the type of subsets to retrieve. Options are ``spatial``
            to retrieve only spatial subsets, ``spectral`` to retrieve only
            spectral subsets or ``None`` (default) to retrieve both spatial
            and spectral subsets, when relevent to the current configuration.

        list_of_subset_labels : list of str or None, optional
            If specified, only subsets matching these labels will be included.
            If not specified, all subsets matching the ``region_type`` will be
            returned.

        use_display_units : bool, optional
            (For spectral subsets) If False (default), subsets are returned in
            the native data unit. If True, subsets are returned in the spectral
            axis display unit set in the Unit Conversion plugin.

        Returns
        -------
        regions : dict
            A dictionary mapping subset labels to their respective ``regions``
            objects (for spatial regions) or ``SpectralRegions`` objects
            (for spectral regions).
        """

        if region_type is not None:
            region_type = region_type.lower()
            if region_type not in ['spectral', 'spatial']:
                raise ValueError("`region_type` must be 'spectral', 'spatial', or None for any.")
            if ((self.config == 'imviz' and region_type == 'spectral') or
               (self.config == 'specviz' and region_type == 'spatial')):
                raise ValueError(f"No {region_type} subests in {self.config}.")
            region_type = [region_type]

        else:  # determine subset return type(s) by config, if not specified
            region_type = {'imviz': ['spatial'],
                           'specviz': ['spectral']}.get(self.config, ['spatial', 'spectral'])

        reg_type = 'sky_region' if self.app._align_by == 'wcs' else 'region'

        # first get ALL subsets of specified spatial/spectral type(s)
        subsets = self.app.get_subsets(spectral_only=region_type == ['spectral'],
                                       spatial_only=region_type == ['spatial'],
                                       include_sky_region=reg_type == 'sky_region',
                                       use_display_units=use_display_units)

        labels = list_of_subset_labels or list(subsets.keys())
        if isinstance(labels, str):
            labels = [labels]

        regions, failed_regs = {}, set()
        for subset_label in labels:
            try:
                ss = subsets[subset_label]
                if isinstance(ss, SpectralRegion):
                    regions[subset_label] = ss
                else:
                    reg = _chain_regions([x[reg_type] for x in ss],
                                         [x['glue_state'] for x in ss])
                    if reg is None:
                        failed_regs.add(subset_label)
                    else:
                        regions[subset_label] = reg
            except ValueError:  # key doesnt exist, subset wasn't retrieved
                failed_regs.add(subset_label)

        if len(failed_regs) > 0:
            warn_msg = f"Regions skipped: {', '.join(sorted(failed_regs))}"
            self.app.hub.broadcast(SnackbarMessage(warn_msg,
                                   color="warning", timeout=8000,
                                   sender=self.app))
            warnings.warn(warn_msg, UserWarning)

        return regions

    def _on_link_update(self, *args):
        """When linking is changed pixels<>wcs, change display units of the
        subset plugin from pixel (for pixel linking) to sky (for WCS linking).
        If there is an active selection in the subset plugin, push this change
        to the UI upon link change by calling _get_subset_definition, which
        will re-determine how to display subset information."""

        align_by = getattr(self.app, '_align_by', None)
        self.display_sky_coordinates = (align_by == 'wcs')

        if self.subset_selected != self.subset.default_text:
            self._get_subset_definition(*args)

    def _sync_selected_from_state(self, *args):
        if not hasattr(self, 'subset') or self.multiselect:
            # during initial init, this can trigger before the component is initialized
            return
        if self.session.edit_subset_mode.edit_subset == []:
            if self.subset_selected != self.subset.default_text:
                self.subset_selected = self.subset.default_text
                self.show_region_info = False
        else:
            new_label = self.session.edit_subset_mode.edit_subset[0].label
            if new_label != self.subset_selected:
                if new_label not in [s['label'] for s in self.subset_items]:
                    self._sync_available_from_state()
                self.subset_selected = self.session.edit_subset_mode.edit_subset[0].label
                self.show_region_info = True
        self._update_combination_mode()

    def _on_subset_update(self, msg):
        self._sync_selected_from_state()
        if 'Create New' in self.subset_selected:
            return
        subsets_avail = [sg.label for sg in self.app.data_collection.subset_groups]
        if self.subset_selected not in subsets_avail:
            # subset selection should re-default after processing the deleted subset,
            # for now we can safely ignore
            return
        self._get_subset_definition()
        subset_to_update = self.session.edit_subset_mode.edit_subset[0]
        self.subset._update_subset(subset_to_update, attribute="type")

    def _sync_available_from_state(self, *args):
        if not hasattr(self, 'subset'):
            # during initial init, this can trigger before the component is initialized
            return
        self.subset_items = [{'label': self.subset.default_text}] + [
                             self.subset._subset_to_dict(subset) for subset in
                             self.data_collection.subset_groups]

    @observe('subset_selected')
    def _sync_selected_from_ui(self, change):
        self.subset_definitions = []
        self.subset_types = []
        self.glue_state_types = []
        self.is_centerable = False

        if not hasattr(self, 'subset'):
            # during initial init, this can trigger before the component is initialized
            return
        if change['new'] != self.subset.default_text:
            self._get_subset_definition(change['new'])
        self.show_region_info = change['new'] != self.subset.default_text
        m = [s for s in self.app.data_collection.subset_groups if s.label == change['new']]
        if m != self.session.edit_subset_mode.edit_subset:
            self.session.edit_subset_mode.edit_subset = m

    def _unpack_get_subsets_for_ui(self):
        """
        Convert what app.get_subsets returns into something the UI of this plugin
        can display.
        """
        if self.multiselect:
            self.is_centerable = True
            return

        include_sky_region = bool(self.display_sky_coordinates)
        subset_information = self.app.get_subsets(self.subset_selected,
                                                  simplify_spectral=False,
                                                  use_display_units=True,
                                                  include_sky_region=include_sky_region)

        _around_decimals = 6  # Avoid 30 degrees from coming back as 29.999999999999996
        if not subset_information:
            return
        if ((len(subset_information) == 1) and
                (isinstance(subset_information[0]["subset_state"], RangeSubsetState) or
                 (isinstance(subset_information[0]["subset_state"], RoiSubsetState) and
                  isinstance(subset_information[0]["subset_state"].roi,
                             (CircularROI, RectangularROI, EllipticalROI))))):
            self.is_centerable = True
        else:
            self.is_centerable = False

        for spec in subset_information:

            subset_definition = []
            subset_type = ''
            subset_state = spec["subset_state"]
            glue_state = spec["glue_state"]

            if isinstance(subset_state, RoiSubsetState):
                subset_definition.append({
                    "name": "Parent", "att": "parent",
                    "value": subset_state.xatt.parent.label,
                    "orig": subset_state.xatt.parent.label})

                sky_region = spec['sky_region']
                if self.display_sky_coordinates and (sky_region is not None):
                    subset_definition += utils._sky_region_to_subset_def(sky_region)

                else:
                    if isinstance(subset_state.roi, CircularROI):
                        x, y = subset_state.roi.center()
                        r = subset_state.roi.radius
                        subset_definition += [
                            {"name": "X Center (pixels)", "att": "xc",
                             "value": x, "orig": x},
                            {"name": "Y Center (pixels)", "att": "yc",
                             "value": y, "orig": y},
                            {"name": "Radius (pixels)", "att": "radius",
                             "value": r, "orig": r}]

                    elif isinstance(subset_state.roi, RectangularROI):
                        for att in ("Xmin", "Xmax", "Ymin", "Ymax"):
                            real_att = att.lower()
                            val = getattr(subset_state.roi, real_att)
                            subset_definition.append(
                                {"name": att + " (pixels)", "att": real_att,
                                 "value": val, "orig": val})

                        theta = np.around(np.degrees(subset_state.roi.theta),
                                          decimals=_around_decimals)
                        subset_definition.append({"name": "Angle", "att": "theta",
                                                  "value": theta, "orig": theta})

                    elif isinstance(subset_state.roi, EllipticalROI):
                        xc, yc = subset_state.roi.center()
                        rx = subset_state.roi.radius_x
                        ry = subset_state.roi.radius_y
                        theta = np.around(np.degrees(subset_state.roi.theta),
                                          decimals=_around_decimals)

                        subset_definition += [
                            {"name": "X Center (pixels)", "att": "xc",
                             "value": xc, "orig": xc},
                            {"name": "Y Center (pixels)", "att": "yc",
                             "value": yc, "orig": yc},
                            {"name": "X Radius (pixels)", "att": "radius_x",
                             "value": rx, "orig": rx},
                            {"name": "Y Radius (pixels)", "att": "radius_y",
                             "value": ry, "orig": ry},
                            {"name": "Angle", "att": "theta",
                             "value": theta, "orig": theta}]

                    elif isinstance(subset_state.roi, CircularAnnulusROI):
                        xc, yc = subset_state.roi.center()
                        inner_r = subset_state.roi.inner_radius
                        outer_r = subset_state.roi.outer_radius
                        subset_definition += [{"name": "X Center (pixels)",
                                               "att": "xc", "value": xc, "orig": xc},
                                              {"name": "Y Center (pixels)",
                                               "att": "yc", "value": yc, "orig": yc},
                                              {"name": "Inner Radius (pixels)",
                                               "att": "inner_radius",
                                               "value": inner_r, "orig": inner_r},
                                              {"name": "Outer Radius (pixels)",
                                               "att": "outer_radius",
                                               "value": outer_r, "orig": outer_r}]

                    else:  # pragma: no cover
                        raise NotImplementedError(f"Unable to translate {subset_state.roi.__class__.__name__}")  # noqa: E501

                subset_type = subset_state.roi.__class__.__name__

            elif isinstance(subset_state, RangeSubsetState):
                region = spec['region']
                if isinstance(region, Time):
                    lo = region.min()
                    hi = region.max()
                    subset_definition = [{"name": "Lower bound", "att": "lo",
                                          "value": lo.value, "orig": lo.value},
                                         {"name": "Upper bound", "att": "hi",
                                          "value": hi.value, "orig": hi.value}]
                else:
                    lo = region.lower
                    hi = region.upper
                    subset_definition = [{"name": "Lower bound", "att": "lo", "value": lo.value,
                                          "orig": lo.value, "unit": str(lo.unit)},
                                         {"name": "Upper bound", "att": "hi", "value": hi.value,
                                          "orig": hi.value, "unit": str(hi.unit)}]
                subset_type = "Range"

            elif isinstance(subset_state, MultiMaskSubsetState):
                total_masked = subset_state.total_masked_first_data()
                subset_definition = [{"name": "Masked values", "att": "masked",
                                      "value": total_masked,
                                      "orig": total_masked}]
                subset_type = "Mask"

            if len(subset_definition) > 0:
                # Note: .append() does not work for List traitlet.
                self.subset_definitions = self.subset_definitions + [subset_definition]
                self.subset_types = self.subset_types + [subset_type]
                self.glue_state_types = self.glue_state_types + [glue_state]
                self.subset_states = self.subset_states + [subset_state]

        simplifiable_states = set(['AndState', 'XorState', 'AndNotState'])
        # Check if the subset has more than one subregion, is a range subset
        # type, and uses one of the states that can be simplified. Mask subset
        # types cannot be simplified so subsets contained that are skipped.
        if 'Mask' in self.subset_types:
            self.can_simplify = False
        elif ((len(self.subset_states) > 1) and isinstance(self.subset_states[0], RangeSubsetState)
              and ((len(simplifiable_states - set(self.glue_state_types)) < 3)
                   or self.app.is_there_overlap_spectral_subset(self.subset_selected))):
            self.can_simplify = True
        else:
            self.can_simplify = False

    def _get_subset_definition(self, *args):
        """
        Retrieve the parameters defining the selected subset, for example the
        upper and lower bounds for a simple spectral subset.
        """
        self.subset_definitions = []
        self.subset_types = []
        self.glue_state_types = []
        self.subset_states = []

        self._unpack_get_subsets_for_ui()

    def vue_freeze_subset(self, *args):
        sgs = {sg.label: sg for sg in self.app.data_collection.subset_groups}
        sg = sgs.get(self.subset_selected)

        masks = {}
        for data in self.app.data_collection:
            masks[data.uuid] = sg.subset_state.to_mask(data)

        sg.subset_state = MultiMaskSubsetState(masks)

    def vue_simplify_subset(self, *args):
        if self.multiselect:
            self.hub.broadcast(SnackbarMessage("Cannot simplify spectral subset "
                                               "when multiselect is active", color='warning',
                                               sender=self))
            return
        if len(self.subset_states) < 2:
            self.hub.broadcast(SnackbarMessage("Cannot simplify spectral subset "
                                               "of length less than 2", color='warning',
                                               sender=self))
            return
        att = self.subset_states[0].att
        self.app.simplify_spectral_subset(subset_name=self.subset_selected, att=att,
                                          overwrite=True)

    def _on_display_unit_changed(self, msg):
        # We only care about the spectral units, since flux units don't affect spectral subsets
        if msg.axis == "spectral":
            self.spectral_display_unit = msg.unit
            if self.subset_selected != self.subset.default_text:
                self._get_subset_definition(self.subset_selected)

    def _update_subset(self):
        status, reason = self._check_input()
        if not status:
            self.hub.broadcast(SnackbarMessage(reason, color='error', sender=self))
            return

        for index, sub in enumerate(self.subset_definitions):
            if len(self.subset_states) <= index:
                return
            sub_states = self.subset_states[index]

            # we need to push updates to subset in pixels. to do this when wcs
            # linked, convert the updated subset parameters from sky to pix
            wcs = None

            if self.display_sky_coordinates:
                wcs = self.app._get_wcs_from_subset(sub_states)

                if wcs is not None:
                    # convert newly entered sky coords to pixel
                    updated_skyreg = utils._subset_def_to_region(self.subset_types[index], sub)  # noqa
                    updated_pixreg_attrs = utils._get_pixregion_params_in_dict(updated_skyreg.to_pixel(wcs))  # noqa
                    # convert previous entered sky coords to pixel
                    orig_skyreg = utils._subset_def_to_region(self.subset_types[index], sub, val='orig')  # noqa
                    orig_pixreg_attrs = utils._get_pixregion_params_in_dict(orig_skyreg.to_pixel(wcs))  # noqa

            for d_att in sub:
                if d_att["att"] == 'parent':  # Read-only
                    continue
                if self.display_sky_coordinates and (wcs is not None):
                    d_att["value"] = updated_pixreg_attrs[d_att["att"]]
                    d_att["orig"] = orig_pixreg_attrs[d_att["att"]]

                if (d_att["att"] == 'theta') and (self.display_sky_coordinates is False):
                    # Humans use degrees but glue uses radians
                    # We've already enforced this in wcs linking in _get_pixregion_params_in_dict
                    d_val = np.radians(d_att["value"])
                else:
                    d_val = float(d_att["value"])

                # Convert from display unit to original unit if necessary
                if self.subset_types[index] == "Range":
                    if self.spectral_display_unit is not None:
                        x_att = sub_states.att
                        # since this is a spectrum range subset, we can get the native units
                        # from the current reference data in the spectrum viewer
                        sv = self.spectrum_viewer
                        base_units = sv.state.reference_data.get_component(x_att).units
                        if self.spectral_display_unit != base_units:
                            d_val = d_val*u.Unit(self.spectral_display_unit)
                            d_val = d_val.to(u.Unit(base_units))
                            d_val = d_val.value

                if float(d_att["orig"]) != d_val:
                    if self.subset_types[index] == "Range":
                        setattr(sub_states, d_att["att"], d_val)
                    else:
                        setattr(sub_states.roi, d_att["att"], d_val)

        self._push_update_to_ui()

    def update_subset(self, subset_label=None, subregion=None, **kwargs):
        '''
        Method to update the attributes of an existing subset. The attributes of a subset
        and their current values can be retrieved with the 'get_subset_definition` method.

        Parameters
        ----------

        subset_label : str

            The name of the subset to update. If this is not the currently selected subset in the
            UI, it will be selected.

        subregion : int, optional

            The integer subregion index (in the subset_definitions dictionary) for which to modify
            the specified attributes.

        The attributes to update and their new values are passed as keyword arguments to this
        function, for example:

            plg = imviz.plugins['Subset Tools']
            plg.update_subset('Subset 1', xmax = 9.522, xmin = 9.452)

        If no values to update are specified, this function will return the current definition of
        the specified subset. The "att" keys in the returned dictionaries are the attributes that
        can be updated with this method.
        '''
        if subset_label is not None:
            if subset_label not in self.subset.choices:
                raise ValueError(f"{subset_label} is not an existing subset. "
                                 f"Available choices are: {self.subset.choices}")

            if subset_label != self.subset.selected:
                self.subset.selected = subset_label

        if not kwargs:
            # If no updates were requested, we instead return the current definition
            public_definition = {}
            for i in range(len(self.subset_definitions)):
                public_definition[f'subregion {i}'] = []
                for d in self.subset_definitions[i]:
                    if d['att'] == 'parent':
                        continue
                    public_definition[f'subregion {i}'].append({key:d[key] for key in d if key != "orig"})  # noqa

            return public_definition

        if len(self.subset_definitions) == 1:
            subregion = 0
        elif subregion is None:
            raise ValueError("Specified subset has more than one subregion, please "
                             "specify which integer subregion index to modify.")

        for key, value in kwargs.items():
            for i in range(len(self.subset_definitions[subregion])):
                att_dict = self.subset_definitions[subregion][i]
                if att_dict['att'] == key:
                    att_dict['value'] = value
                    self.subset_definitions[subregion][i] = att_dict
                    break
            else:
                raise ValueError(f"{key} is not an attribute of the specified subset/subregion.")

        self._update_subset()

    def vue_update_subset(self, *args):

        if self.multiselect:
            self.hub.broadcast(SnackbarMessage("Cannot update subset "
                                               "when multiselect is active", color='warning',
                                               sender=self))
            return

        self._update_subset()

    def _push_update_to_ui(self, subset_name=None):
        """
        Forces the UI to update how it represents the subset.

        Parameters
        ----------
        subset_name : str
            The name of the subset that is being updated.

        """
        if not subset_name:
            subset_name = self.subset_selected
        try:
            dc = self.data_collection
            subsets = dc.subset_groups
            subset_to_update = subsets[[x.label for x in subsets].index(subset_name)]
            self.session.edit_subset_mode.edit_subset = [subset_to_update]
            self.session.edit_subset_mode._combine_data(subset_to_update.subset_state,
                                                        override_mode=ReplaceMode)
        except Exception as err:  # pragma: no cover
            self.hub.broadcast(SnackbarMessage(
                f"Failed to update Subset: {repr(err)}", color='error', sender=self))

    def _check_input(self):
        status = True
        reason = ""
        for index, sub in enumerate(self.subset_definitions):
            lo = hi = xmin = xmax = ymin = ymax = None
            inner_radius = outer_radius = None
            for d_att in sub:
                if d_att["att"] == "lo":
                    lo = d_att["value"]
                elif d_att["att"] == "hi":
                    hi = d_att["value"]
                elif d_att["att"] == "radius" and d_att["value"] <= 0:
                    status = False
                    reason = "Failed to update Subset: radius must be a positive scalar"
                    break
                elif d_att["att"] == "xmin":
                    xmin = d_att["value"]
                elif d_att["att"] == "xmax":
                    xmax = d_att["value"]
                elif d_att["att"] == "ymin":
                    ymin = d_att["value"]
                elif d_att["att"] == "ymax":
                    ymax = d_att["value"]
                elif d_att["att"] == "outer_radius":
                    outer_radius = d_att["value"]
                elif d_att["att"] == "inner_radius":
                    inner_radius = d_att["value"]

                if lo and hi and hi <= lo:
                    status = False
                    reason = "Failed to update Subset: lower bound must be less than upper bound"
                    break
                elif xmin and xmax and ymin and ymax and (xmax - xmin <= 0 or ymax - ymin <= 0):
                    status = False
                    reason = "Failed to update Subset: width and length must be positive scalars"
                    break
                elif inner_radius and outer_radius and inner_radius >= outer_radius:
                    status = False
                    reason = "Failed to update Subset: inner radius must be less than outer radius"
                    break

        return status, reason

    def recenter(self):
        """
        Recenter the selected subset on the centroid of the region of the current subset
        in the selected data layer.
        """
        # Composite region cannot be centered. This only works for Imviz.
        if not self.is_centerable or self.config != 'imviz':  # no-op
            raise NotImplementedError(
                f'Cannot recenter: is_centerable={self.is_centerable}, config={self.config}')

        from astropy.wcs.utils import pixel_to_pixel
        from photutils.aperture import ApertureStats
        from jdaviz.core.region_translators import regions2aperture, _get_region_from_spatial_subset

        def _do_recentering(subset, subset_state):
            try:
                reg = _get_region_from_spatial_subset(self, subset_state)
                aperture = regions2aperture(reg)
                data = self.recenter_dataset.selected_dc_item
                comp = data.get_component(data.main_components[0])
                comp_data = comp.data
                phot_aperstats = ApertureStats(comp_data, aperture, wcs=data.coords)

                # Sky region from WCS linking, need to convert centroid back to pixels.
                if hasattr(reg, "to_pixel"):
                    # Centroid was calculated in selected data.
                    # However, Subset is always defined w.r.t. its parent,
                    # so we need to convert back.
                    x, y = pixel_to_pixel(
                        data.coords,
                        subset_state.xatt.parent.coords,
                        phot_aperstats.xcentroid,
                        phot_aperstats.ycentroid)
                else:
                    x = phot_aperstats.xcentroid
                    y = phot_aperstats.ycentroid
                if not np.all(np.isfinite((x, y))):
                    raise ValueError(f'Invalid centroid ({x}, {y})')
            except Exception as err:
                self.set_center(self.get_center(subset_name=subset), subset_name=subset,
                                update=False)
                self.hub.broadcast(SnackbarMessage(
                    f"Failed to calculate centroid: {repr(err)}", color='error', sender=self))
            else:
                self.set_center((x, y), subset_name=subset, update=True)

        if not self.multiselect:
            _do_recentering(self.subset_selected, self.subset.selected_subset_state)
        else:
            for sub, subset_state in zip(self.subset_selected,
                                         self.subset.selected_subset_state):
                if (sub != self.subset.default_text and
                        not isinstance(subset_state, CompositeSubsetState)):
                    self.is_centerable = True
                    _do_recentering(sub, subset_state)
                elif (sub != self.subset.default_text and
                      isinstance(subset_state, CompositeSubsetState)):
                    self.hub.broadcast(SnackbarMessage(f"Unable to recenter "
                                                       f"composite subset {sub}",
                                                       color='error', sender=self))

    def vue_recenter_subset(self, *args):
        self.recenter()

    def _get_subset_state(self, subset_name=None):
        if self.multiselect and not subset_name:
            raise ValueError("Please include subset_name in when in multiselect mode")

        if subset_name is not None:
            return self.subset._get_subset_state(subset_name)
        # guaranteed to only return a single entry because of check above
        return self.subset.selected_subset_state

    def get_center(self, subset_name=None):
        """Return the center of the Subset.
        This may or may not be the centroid obtain from data.

        Parameters
        ----------
        subset_name : str
            The name of the subset that is being updated.

        Returns
        -------
        cen : number, tuple of numbers, or `None`
            The center of the Subset in ``x`` or ``(x, y)``,
            depending on the Subset type, if applicable.
            If Subset is not centerable, this returns `None`.

        """
        # Composite region cannot be centered.
        if not self.is_centerable:  # no-op
            return

        subset_state = self._get_subset_state(subset_name)
        return subset_state.center()

    def set_center(self, new_cen, subset_name=None, update=False):
        """Set the desired center for the selected Subset, if applicable.
        If Subset is not centerable, nothing is done.

        Parameters
        ----------
        new_cen : number or tuple of numbers
            The new center defined either as ``x`` or ``(x, y)``,
            depending on the Subset type.
        subset_name : str
            The name of the subset that is being updated.
        update : bool
            If `True`, the Subset is also moved to the new center.
            Otherwise, only the relevant editable fields are updated but the
            Subset is not moved.

        Raises
        ------
        NotImplementedError
            Subset type is not supported.

        """
        # Composite region cannot be centered, so just grab first element.
        if not self.is_centerable:  # no-op
            return

        subset_state = self._get_subset_state(subset_name)

        if isinstance(subset_state, RoiSubsetState):
            x, y = new_cen
            # x and y are arrays so this converts them back to floats
            x = float(x)
            y = float(y)
            sbst_obj = subset_state.roi
            if isinstance(sbst_obj, (CircularROI, CircularAnnulusROI, EllipticalROI, RectangularROI)):  # noqa
                sbst_obj.move_to(x, y)
            else:  # pragma: no cover
                raise NotImplementedError(f'Recentering of {sbst_obj.__class__} is not supported')

        elif isinstance(subset_state, RangeSubsetState):
            subset_state.move_to(new_cen)

        else:  # pragma: no cover
            raise NotImplementedError(
                f'Getting center of {subset_state.__class__} is not supported')

        if update:
            self._push_update_to_ui(subset_name=subset_name)
        else:
            # Force UI to update on browser without changing the subset.
            tmp = self.subset_definitions
            self.subset_definitions = []
            self.subset_definitions = tmp

    # List of JSON-like dict is nice for front-end but a pain to look up,
    # so we use these helper functions.

    def _get_value_from_subset_definition(self, index, name, desired_key):
        subset_definition = self.subset_definitions[index]
        value = None
        for item in subset_definition:
            if item['name'] == name:
                value = item[desired_key]
                break
        return value

    def _set_value_in_subset_definition(self, index, name, desired_key, new_value):
        for i in range(len(self.subset_definitions[index])):
            if self.subset_definitions[index][i]['name'] == name:
                self.subset_definitions[index][i]['value'] = new_value
                break

    @cached_property
    def selected_subset_group(self):
        for subset_group in self.app.data_collection.subset_groups:
            if subset_group.label == self.subset.selected:
                return subset_group

    def rename_subset(self, old_label, new_label):
        """
        Method to rename an existing subset

        Parameters
        ----------
        old_label : str
            The current label of the subset to be renamed.
        new_label : str
            The new label to apply to the selected subset.
        """
        # will emit a SubsetRenameMessage and call _sync_available_from_state()
        self.subset.rename_choice(old_label, new_label)

    def vue_rename_subset(self, msg):
        try:
            self.rename_subset(msg['old_label'], msg['new_label'])
        except Exception as e:
            self.hub.broadcast(SnackbarMessage(f"Failed to rename subset: {repr(e)}",
                                               color='error', sender=self))

    def rename_selected(self, new_label):
        """
        Method to rename the subset currently selected in the Subset Tools plugin.

        Parameters
        ----------
        new_label : str
            The new label to apply to the selected subset.
        """
        subset_group = self.selected_subset_group
        if subset_group is None:
            raise TypeError("current selection is not a subset")

        self.app._rename_subset(self.subset.selected, new_label, subset_group=subset_group)
        self._sync_available_from_state()

    def import_region(self, region, combination_mode=None, max_num_regions=20,
                      refdata_label=None, return_bad_regions=False, region_format=None):
        """
        Method for creating subsets from regions or region files.

        Parameters
        ----------
        region : region, list of region objects, or str
            A region object can be one of the following:

            * Astropy ``regions`` object
            * ``photutils`` apertures (limited support until ``photutils``
              fully supports ``regions``)
            * specutils ``SpectralRegion`` object

            A string which represents a ``regions`` or ``SpectralRegion`` file.
            If given as a list, it can only contain spectral or non-spectral regions, not both.

        combination_mode : list, str, or `None`
            The way that regions are created or combined. If a list, then it must be the
            same length as regions. If `None`, then it will follow the default glue
            functionality for subset creation.

        max_num_regions : int or `None`
            Maximum number of regions to load, starting from top of the list.
            Default is 20.  If you want to load everything, set it to `None`.
            Loading a large number of regions is not recommended due to performance impact.

        refdata_label : str or `None`
            **This is only applicable to non-spectral regions.**
            Label of data to use for sky-to-pixel conversion for a region, or
            mask creation. Data must already be loaded into Jdaviz.
            If `None`, defaults to the reference data in the default viewer.
            Choice of this data is particularly important when sky
            region is involved.

        return_bad_regions : bool
            If `True`, return the regions that failed to load (see ``bad_regions``);
            This is useful for debugging. If `False`, do not return anything (`None`).

        region_format : str or `None`
            Passed to ``Regions.read(format=region_format)``.  Only applicable if ``region``
            is a string pointing to a valid file that ``Regions`` can read.

        Returns
        -------
        bad_regions : list of (obj, str) or `None`
            If requested (see ``return_bad_regions`` option), return a
            list of ``(region, reason)`` tuples for region objects that failed to load.
            If all the regions loaded successfully, this list will be empty.
            If not requested, return `None`.

        """
        if isinstance(region, str):
            if os.path.exists(region):
                from regions import Regions
                try:
                    raw_regs = Regions.read(region, format=region_format)
                except Exception:  # nosec
                    raw_regs = SpectralRegion.read(region)

                return self._load_regions(raw_regs, combination_mode, max_num_regions,
                                          refdata_label, return_bad_regions)
        else:
            return self._load_regions(region, combination_mode, max_num_regions, refdata_label,
                                      return_bad_regions)

    def _load_regions(self, regions, combination_mode=None, max_num_regions=None,
                      refdata_label=None, return_bad_regions=False, **kwargs):
        """Load given region(s) into the viewer.
        WCS-to-pixel translation and mask creation, if needed, is relative
        to the image defined by ``refdata_label``. Meanwhile, the rest of
        the Subset operations are based on reference image as defined by Glue.

        .. note:: Loading too many regions will affect performance.

        A valid region can be loaded into one of the following categories:

        * An interactive Subset, as if it was drawn by hand. This is
          always done for supported shapes. Its label will be
          ``'Subset N'``, where ``N`` is an integer.
        * A masked Subset that will display on the image but cannot be
          modified once loaded. This is done if the shape cannot be
          made interactive. Its label will be ``'MaskedSubset N'``,
          where ``N`` is an integer.

        Parameters
        ----------
        regions : list of obj
            A list of region objects. A region object can be one of the following:

            * Astropy ``regions`` object
            * ``photutils`` apertures (limited support until ``photutils``
              fully supports ``regions``)
            * specutils ``SpectralRegion`` object

        combination_mode : list, str, or `None`
            The way that regions are created or combined. If a list, then it must be the
            same length as regions. Each element describes how the corresponding region
            in regions will be applied. If `None`, then it will follow the default glue
            functionality for subset creation. Options are ['new', 'replace', 'or', 'and',
            'xor', 'andnot']

        max_num_regions : int or `None`
            Maximum number of regions to load, starting from top of the list.
            Default is to load everything.

        refdata_label : str or `None`
            Label of data to use for sky-to-pixel conversion for a region, or
            mask creation. Data must already be loaded into Jdaviz.
            If `None`, defaults to the reference data in the default viewer.
            Choice of this data is particularly important when sky
            region is involved.

        return_bad_regions : bool
            If `True`, return the regions that failed to load (see ``bad_regions``);
            This is useful for debugging. If `False`, do not return anything (`None`).

        kwargs : dict
            Extra keywords to be passed into the region's ``to_mask`` method.
            **This is ignored if the region can be made interactive.**

        Returns
        -------
        bad_regions : list of (obj, str) or `None`
            If requested (see ``return_bad_regions`` option), return a
            list of ``(region, reason)`` tuples for region objects that failed to load.
            If all the regions loaded successfully, this list will be empty.
            If not requested, return `None`.

        """
        if len(self.app.data_collection) == 0:
            raise ValueError('Cannot load regions without data.')

        if not isinstance(regions, (list, tuple, Regions, SpectralRegion)):
            regions = [regions]

        n_loaded = 0
        bad_regions = []

        # To keep track of masked subsets.
        msg_prefix = 'MaskedSubset'
        msg_count = _next_subset_num(msg_prefix, self.app.data_collection.subset_groups)

        viewer_parameter = kwargs.pop('viewer', None)
        viewer_name = viewer_parameter or list(self.app._jdaviz_helper.viewers.keys())[0]
        viewer = self.app.get_viewer(viewer_name)
        # Subset is global but reference data is viewer-dependent.
        if refdata_label is None:
            data = viewer.state.reference_data
        else:
            data = self.app.data_collection[refdata_label]

        has_wcs = data_has_valid_wcs(data, ndim=2)

        combo_mode_is_list = isinstance(combination_mode, list)
        if combo_mode_is_list and len(combination_mode) != (len(regions)):
            raise ValueError("list of mode must be size of regions")
        elif combo_mode_is_list:
            unknown_options = list(set(combination_mode) - set(COMBO_OPTIONS))
            if len(unknown_options) > 0:
                raise ValueError(f"{unknown_options} not one of {COMBO_OPTIONS}")

        previous_mode = self.app.session.edit_subset_mode.mode

        with self.app._jdaviz_helper.batch_load():
            for index, region in enumerate(regions):
                # Set combination mode for how region will be applied to current subset
                # or created as a new subset
                if combo_mode_is_list:
                    combo_mode = combination_mode[index]
                else:
                    combo_mode = combination_mode

                if combo_mode == 'new':
                    # Remove selection of subset so that new one will be created
                    # No overwrite next iteration
                    self.app.session.edit_subset_mode.edit_subset = None
                    self.app.session.edit_subset_mode.mode = SUBSET_MODES_PRETTY['new']
                elif combo_mode:
                    self.combination_mode.selected = combo_mode

                if (isinstance(region, (SkyCircularAperture, SkyEllipticalAperture,
                                        SkyRectangularAperture, SkyCircularAnnulus,
                                        CircleSkyRegion, EllipseSkyRegion,
                                        RectangleSkyRegion, CircleAnnulusSkyRegion))
                        and not has_wcs):
                    bad_regions.append((region, 'Sky region provided but data has no valid WCS'))
                    continue

                if (isinstance(region, (CircularAperture, EllipticalAperture,
                                        RectangularAperture, CircularAnnulus,
                                        CirclePixelRegion, EllipsePixelRegion,
                                        RectanglePixelRegion, CircleAnnulusPixelRegion))
                        and (hasattr(self.app, '_link_type') and self.app._link_type == "wcs")):
                    bad_regions.append((region, 'Pixel region provided by data is linked by WCS'))
                    continue

                # photutils: Convert to region shape first
                if isinstance(region, (CircularAperture, SkyCircularAperture,
                                       EllipticalAperture, SkyEllipticalAperture,
                                       RectangularAperture, SkyRectangularAperture,
                                       CircularAnnulus, SkyCircularAnnulus)):
                    region = aperture2regions(region)

                # region: Convert to ROI.
                # NOTE: Out-of-bounds ROI will succeed; this is native glue behavior.
                if isinstance(region, (CirclePixelRegion, CircleSkyRegion,
                                       EllipsePixelRegion, EllipseSkyRegion,
                                       RectanglePixelRegion, RectangleSkyRegion,
                                       CircleAnnulusPixelRegion, CircleAnnulusSkyRegion)):
                    state = regions2roi(region, wcs=data.coords)
                    viewer.apply_roi(state)

                elif isinstance(region, (CircularROI, CircularAnnulusROI,
                                         EllipticalROI, RectangularROI)):
                    viewer.apply_roi(region)

                elif isinstance(region, SpectralRegion):
                    # Use viewer_name if provided in kwarg, otherwise use
                    # default spectrum viewer name
                    viewer_name = (viewer_parameter or
                                   self.app._jdaviz_helper._default_spectrum_viewer_reference_name)
                    range_viewer = self.app.get_viewer(viewer_name)

                    s = RangeSubsetState(lo=region.lower.value, hi=region.upper.value,
                                         att=range_viewer.state.x_att)
                    range_viewer.apply_subset_state(s)

                # Last resort: Masked Subset that is static (if data is not a cube)
                elif data.ndim == 2:
                    im = None
                    if hasattr(region, 'to_pixel'):  # Sky region: Convert to pixel region
                        if not has_wcs:
                            bad_regions.append((region, 'Sky region provided but data has no valid WCS'))  # noqa
                            continue
                        region = region.to_pixel(data.coords)

                    if hasattr(region, 'to_mask'):
                        try:
                            mask = region.to_mask(**kwargs)
                            im = mask.to_image(data.shape)  # Can be None
                        except Exception as e:  # pragma: no cover
                            bad_regions.append((region, f'Failed to load: {repr(e)}'))
                            continue

                    # Boolean mask as input is supported but not advertised.
                    elif (isinstance(region, np.ndarray) and region.shape == data.shape
                          and region.dtype == np.bool_):
                        im = region

                    if im is None:
                        bad_regions.append((region, 'Mask creation failed'))
                        continue

                    # NOTE: Region creation info is thus lost.
                    try:
                        subset_label = f'{msg_prefix} {msg_count}'
                        state = MaskSubsetState(im, data.pixel_component_ids)
                        self.app.data_collection.new_subset_group(subset_label, state)
                        msg_count += 1
                    except Exception as e:  # pragma: no cover
                        bad_regions.append((region, f'Failed to load: {repr(e)}'))
                        continue
                else:
                    bad_regions.append((region, 'Mask creation failed'))
                    continue
                n_loaded += 1
                if max_num_regions is not None and n_loaded >= max_num_regions:
                    break

        # Revert edit mode to before the import_region call
        self.app.session.edit_subset_mode.mode = previous_mode

        n_reg_in = len(regions)
        n_reg_bad = len(bad_regions)
        if n_loaded == 0:
            snack_color = "error"
        elif n_reg_bad > 0:
            snack_color = "warning"
        else:
            snack_color = "success"
        self.app.hub.broadcast(SnackbarMessage(
            f"Loaded {n_loaded}/{n_reg_in} regions, max_num_regions={max_num_regions}, "
            f"bad={n_reg_bad}", color=snack_color, timeout=8000, sender=self.app))

        if return_bad_regions:
            return bad_regions

    @observe('combination_mode_selected')
    def _combination_mode_selected_updated(self, change):
        self.app.session.edit_subset_mode.mode = SUBSET_MODES_PRETTY[change['new']]

    def _update_combination_mode(self):
        if self.app.session.edit_subset_mode.mode in SUBSET_TO_PRETTY.keys():
            self.combination_mode_selected = SUBSET_TO_PRETTY[
                self.app.session.edit_subset_mode.mode]



# This is adapted from Ginga (ginga.util.wcs, ginga.trcalc, and ginga.Bindings.ImageViewBindings).
# Please see the file licenses/GINGA_LICENSE.txt for details.
#
"""This module handles calculations based on world coordinate system (WCS)."""

import base64
import math
from io import BytesIO

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from astropy import coordinates as coord
from astropy.coordinates import SkyCoord
from astropy.modeling import models
from astropy.nddata import NDData
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales

from gwcs import coordinate_frames as cf
from gwcs.wcs import WCS as GWCS

from matplotlib.patches import Polygon
from jdaviz.utils import _wcs_only_label

__all__ = ['get_compass_info', 'draw_compass_mpl']


def rotate_pt(x_arr, y_arr, theta_deg, xoff=0, yoff=0):
    """
    Rotate an array of points (x_arr, y_arr) by theta_deg offsetted
    from a center point by (xoff, yoff).
    """
    a_arr = x_arr - xoff
    b_arr = y_arr - yoff
    cos_t = np.cos(np.radians(theta_deg))
    sin_t = np.sin(np.radians(theta_deg))
    ap = (a_arr * cos_t) - (b_arr * sin_t)
    bp = (a_arr * sin_t) + (b_arr * cos_t)
    return np.asarray((ap + xoff, bp + yoff))


def add_offset_radec(ra_deg, dec_deg, delta_deg_ra, delta_deg_dec):
    """
    Algorithm to compute RA/Dec from RA/Dec base position plus tangent
    plane offsets.
    """
    # To radians
    x = math.radians(delta_deg_ra)
    y = math.radians(delta_deg_dec)
    raz = math.radians(ra_deg)
    decz = math.radians(dec_deg)

    sdecz = math.sin(decz)
    cdecz = math.cos(decz)

    d = cdecz - y * sdecz

    ra2 = math.atan2(x, d) + raz
    # Normalize ra into the range 0 to 2*pi
    twopi = math.pi * 2
    ra2 = math.fmod(ra2, twopi)
    if ra2 < 0.0:
        ra2 += twopi
    dec2 = math.atan2(sdecz + y * cdecz, math.sqrt(x * x + d * d))

    # back to degrees
    ra2_deg = math.degrees(ra2)
    dec2_deg = math.degrees(dec2)

    return (ra2_deg, dec2_deg)


def add_offset_xy(image_wcs, x, y, delta_deg_x, delta_deg_y):
    # calculate ra/dec of x,y pixel
    c = image_wcs.pixel_to_world(x, y)
    if isinstance(c, SkyCoord):
        ra_deg = c.ra.deg
        dec_deg = c.dec.deg
    else:  # list of Quantity (e.g., from FITS primary header)
        ra_deg = c[0].value
        dec_deg = c[1].value

    # add offsets
    ra2_deg, dec2_deg = add_offset_radec(ra_deg, dec_deg, delta_deg_x, delta_deg_y)

    # then back to new pixel coords
    return image_wcs.world_to_pixel_values(ra2_deg, dec2_deg)  # x2, y2


def calc_compass(image_wcs, x, y, len_deg_e, len_deg_n):

    # Get east and north coordinates
    xe, ye = list(map(float, add_offset_xy(image_wcs, x, y, len_deg_e, 0.0)))
    xn, yn = list(map(float, add_offset_xy(image_wcs, x, y, 0.0, len_deg_n)))

    return (x, y, xn, yn, xe, ye)


def calc_compass_radius(image_wcs, x, y, radius_px):

    # Define an angular length we can use to determine the pixel scale
    # along east and north - we want to make sure we use a small length so
    # that the offset points still fall inside the image in case they have
    # a bounding box set.

    delta = 0.1 / 3600  # 0.1 arcsec

    xe, ye = add_offset_xy(image_wcs, x, y, delta, 0.0)
    xn, yn = add_offset_xy(image_wcs, x, y, 0.0, delta)

    # now calculate the length in pixels of those arcs
    # (planar geometry is good enough here)
    px_per_delta_e = math.sqrt(math.fabs(ye - y) ** 2 + math.fabs(xe - x) ** 2)
    px_per_delta_n = math.sqrt(math.fabs(yn - y) ** 2 + math.fabs(xn - x) ** 2)

    # now calculate the arm length in degrees for each arm
    # (this produces same-length arms)
    len_deg_e = radius_px / px_per_delta_e * delta
    len_deg_n = radius_px / px_per_delta_n * delta

    return calc_compass(image_wcs, x, y, len_deg_e, len_deg_n)


def calc_compass_center(image_wcs, image_shape, r_fac=0.5):
    # calculate center of data
    x = image_shape[1] * 0.5
    y = image_shape[0] * 0.5

    # radius we want the arms to be
    radius_px = min(image_shape) * r_fac

    return calc_compass_radius(image_wcs, x, y, radius_px)


def get_compass_info(image_wcs, image_shape, r_fac=0.4):
    """Calculate WCS compass parameters.
    North (N) is up and East (E) is left.

    Parameters
    ----------
    image_wcs : obj
        WCS that is compatible with APE 14.

    image_shape : tuple of int
        Shape of the image in the form of ``(ny, nx)``.

    r_fac : float
        Scale factor for compass arrow length.

    Returns
    -------
    x, y : float
        Pixel positions for the center of the compass.

    xn, yn : float
        Pixel positions for N of the compass.

    xe, ye : float
        Pixel positions for E of the compass.

    degn, dege : float
        Rotation angles for N and E, in degrees, for the compass, respectively.

    xflip : bool
        Should display flip on X?

    """
    x, y, xn, yn, xe, ye = calc_compass_center(image_wcs, image_shape, r_fac=r_fac)
    degn = math.degrees(math.atan2(xn - x, yn - y))

    # rotate east point also by degn
    xe2, ye2 = rotate_pt(xe, ye, degn, xoff=x, yoff=y)
    dege = math.degrees(math.atan2(xe2 - x, ye2 - y))

    # if right-hand image, flip it to make left hand
    xflip = False
    if dege > 0.0:
        xflip = not xflip
    if xflip and not np.isclose(degn, 0):
        degn = -degn

    return x, y, xn, yn, xe, ye, degn, dege, xflip


def draw_compass_mpl(image, orig_shape=None, wcs=None, show=True, zoom_limits=None, **kwargs):
    """Visualize the compass using Matplotlib.

    Parameters
    ----------
    image : ndarray
        2D Numpy array (can be resampled).

    orig_shape : tuple of int or `None`
        The original (non-resampled) array shape in ``(ny, nx)``, if different.

    wcs : obj or `None`
        Associated original image WCS that is compatible with APE 14.
        If `None` given, compass is not drawn.

    show : bool
        Display the plot.

    zoom_limits : ndarray or None
        If not `None`, also draw a rectangle to represent the
        current zoom limits in the form of list of ``(x, y)``
        representing the four corners of the zoom box.

    kwargs : dict
        Keywords for ``matplotlib.pyplot.imshow``.

    Returns
    -------
    image_base64 : str
        Decoded buffer for Compass plugin.

    """
    if orig_shape is None:
        orig_shape = image.shape

    if not show:
        plt.ioff()

    fig, ax = plt.subplots()
    ax.imshow(image, extent=[-0.5, orig_shape[1] - 0.5, -0.5, orig_shape[0] - 0.5],
              origin='lower', cmap='gray', **kwargs)

    if wcs is not None:
        try:
            x, y, xn, yn, xe, ye, degn, dege, xflip = get_compass_info(wcs, orig_shape)
        except Exception:
            wcs = None
        else:
            # TODO: Not sure what xflip really do, ask Eric Jeschke later.
            # if xflip:
            #    plt.imshow(np.fliplr(image), origin='lower')

            # Positive here is counter-clockwise, hence the minus sign in comment.
            ax.plot(x, y, marker='o', color='cyan', markersize=5)
            ax.annotate('N', xy=(x, y), xytext=(xn, yn),
                        arrowprops={'arrowstyle': '<-', 'color': 'cyan', 'lw': 1.5},
                        color='cyan', fontsize=16, va='center', ha='center')  # rotation=-degn
            ax.annotate('E', xy=(x, y), xytext=(xe, ye),
                        arrowprops={'arrowstyle': '<-', 'color': 'cyan', 'lw': 1.5},
                        color='cyan', fontsize=16, va='center', ha='center')  # rotation=-dege
    if wcs is None:
        x = orig_shape[1] * 0.5
        y = orig_shape[0] * 0.5
        ax.plot(x, y, marker='o', color='yellow', markersize=5)

    # Also draw X/Y compass.
    r_xy = float(min(orig_shape)) * 0.25
    ax.annotate('X', xy=(x, y), xytext=(x + r_xy, y),
                arrowprops={'arrowstyle': '<-', 'color': 'yellow', 'lw': 1.5},
                color='yellow', fontsize=16, va='center', ha='center')
    ax.annotate('Y', xy=(x, y), xytext=(x, y + r_xy),
                arrowprops={'arrowstyle': '<-', 'color': 'yellow', 'lw': 1.5},
                color='yellow', fontsize=16, va='center', ha='center')

    if zoom_limits is not None:
        ax.add_patch(Polygon(
            zoom_limits, closed=True, linewidth=1.5, edgecolor='r', facecolor='none'))

    if show:
        plt.draw()
        plt.show()

    buff = BytesIO()
    plt.savefig(buff)
    plt.style.use('default')
    plt.close()

    return base64.b64encode(buff.getvalue()).decode('utf-8')


def data_outside_gwcs_bounding_box(data, x, y):
    """This is for internal use by Imviz coordinates transformation only."""
    outside_bounding_box = False
    if hasattr(data.coords, '_orig_bounding_box'):
        # then coords is a GWCS object and had its bounding box cleared
        # by the Imviz parser
        ints = data.coords._orig_bounding_box.intervals
        if isinstance(ints[0].lower, u.Quantity):
            bb_xmin = ints[0].lower.value
            bb_xmax = ints[0].upper.value
            bb_ymin = ints[1].lower.value
            bb_ymax = ints[1].upper.value
        else:  # pragma: no cover
            bb_xmin = ints[0].lower
            bb_xmax = ints[0].upper
            bb_ymin = ints[1].lower
            bb_ymax = ints[1].upper
        if not (bb_xmin <= x <= bb_xmax and bb_ymin <= y <= bb_ymax):
            outside_bounding_box = True  # Has to be Python bool, not Numpy bool_
    return outside_bounding_box


def _rotated_gwcs(
    center_world_coord,
    rotation_angle,
    pixel_scales,
    cdelt_signs,
    refdata_shape=(10, 10),
    image_shape=None
):
    # based on ``gwcs_simple_imaging_units`` in gwcs:
    #   https://github.com/spacetelescope/gwcs/blob/
    #   eec9a2b6de8356495f405de3dc6531538589ce5d/gwcs/tests/conftest.py#L165
    image_extent = u.Quantity(image_shape, u.pix) * u.Quantity(pixel_scales)
    refdata_extent = image_extent.max()
    pixel_scales = refdata_extent / u.Quantity(refdata_shape, u.pix)

    # multiplying by +/-1 can flip north/south or east/west:
    flip_direction = (
        models.Multiply(cdelt_signs[0]) &
        models.Multiply(cdelt_signs[1])
    )

    # shift to compensate for the difference between the center and corner:
    shift = (
        models.Shift((0.5 - refdata_shape[0])/2 * u.pix) &
        models.Shift((0.5 - refdata_shape[1])/2 * u.pix)
    )

    # rotate field of view:
    rho = rotation_angle
    sin_rho = np.sin(rho.to_value(u.rad))
    cos_rho = np.cos(rho.to_value(u.rad))
    rotation_matrix = np.array([[cos_rho, -sin_rho],
                                [sin_rho, cos_rho]])
    rotation = models.AffineTransformation2D(
        rotation_matrix * u.deg, translation=[0, 0] * u.deg
    )
    rotation.input_units_equivalencies = {
        "x": u.pixel_scale(pixel_scales[0]),
        "y": u.pixel_scale(pixel_scales[1])
    }
    rotation.inverse = models.AffineTransformation2D(
        np.linalg.inv(rotation_matrix) * u.pix, translation=[0, 0] * u.pix
    )
    rotation.inverse.input_units_equivalencies = {
        "x": u.pixel_scale(1 / pixel_scales[0]),
        "y": u.pixel_scale(1 / pixel_scales[1])
    }

    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        center_world_coord.ra, center_world_coord.dec, 180 * u.deg
    )

    det2sky = shift | flip_direction | rotation | tan | celestial_rotation
    det2sky.name = "linear_transform"

    detector_frame = cf.Frame2D(
        name="detector",
        axes_names=("x", "y"),
        unit=(u.pix, u.pix)
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.ICRS(),
        name='icrs',
        unit=(u.deg, u.deg)
    )
    pipeline = [
        (detector_frame, det2sky),
        (sky_frame, None)
    ]

    return GWCS(pipeline)


def _prepare_rotated_nddata(real_image_shape, wcs, rotation_angle, refdata_shape,
                            wcs_only_key="_WCS_ONLY", data=None,
                            cdelt_signs=None):
    cdelt = None
    # compute the x/y pixel scales from the WCS:
    if hasattr(wcs, 'pixel_scale_matrix'):
        pixel_scales = u.Quantity([
            value * (unit / u.pix)
            for value, unit in zip(
                proj_plane_pixel_scales(wcs), wcs.wcs.cunit
            )
        ])
        if getattr(wcs.wcs, 'cd', None) is not None:
            cdelt = np.diag(wcs.wcs.cd)
        else:
            cdelt = wcs.wcs.cdelt

    elif data.meta.get(wcs_only_key, False):
        # WCS-only layers have pixel scales in meta:
        pixel_scales = u.Quantity(data.meta['_pixel_scales'])

    elif 'wcsinfo' in data.meta and 'wcs' in data.meta and 'ra_ref' in data.meta['wcsinfo']:
        # GWCS doesn't yet have a pixel scale attr, so approximate
        # its behavior using the pixel scale method from jwst:
        pixel_scales = (2 * [compute_scale(
            data.meta['wcs'],
            (data.meta['wcsinfo']['ra_ref'],
             data.meta['wcsinfo']['dec_ref']),
            1
        )]) * u.deg / u.pix
    else:
        # fall back on CRVAL cards if they're available
        wcsinfo = (
            data.meta.get('_primary_header', None) or
            data.meta.get('wcsinfo', None) or
            data.meta.get('wcs', None)
        )
        if wcsinfo is not None and not isinstance(wcsinfo, GWCS):
            crval1 = float(wcsinfo.get('CRVAL1', wcsinfo.get('crval1')))
            crval2 = float(wcsinfo.get('CRVAL2', wcsinfo.get('crval2')))
            cdelt = [
                float(wcsinfo.get('CDELT1', wcsinfo.get('cdelt1'))),
                float(wcsinfo.get('CDELT2', wcsinfo.get('cdelt2')))
            ]
            unit = u.Unit(wcsinfo.get('CUNIT1', wcsinfo.get('cunit1')))
            fiducial = [crval1, crval2] * unit
            pixel_scales = (2 * [compute_scale(
                WCS(data.meta['_primary_header'])
                if 'wcs' not in data.meta else data.meta['wcs'],
                fiducial, None, 1
            )]) * u.deg / u.pix
        else:
            # fall back on simple approximation:
            compare_pixel_coords = [[0, 0], [0, 1]] * u.pix
            compare_sky_coords = data.coords.pixel_to_world(*compare_pixel_coords)
            separation = compare_sky_coords[0].separation(compare_sky_coords[1])
            pixel_scales = u.Quantity([separation, separation]) / u.pix

    # flip e.g. RA or Dec axes?
    if cdelt_signs is None and cdelt is not None:
        cdelt_signs = np.sign(cdelt)

    # get the world coordinates of the pixel origin
    center_pixel_coord = np.array(real_image_shape) / 2 * u.pix
    center_world_coord = wcs.pixel_to_world(*center_pixel_coord[::-1])
    rotation_angle = coord.Angle(rotation_angle).wrap_at(360 * u.deg)

    # create a GWCS centered on ``filename``,
    # and rotated by ``rotation_angle``:
    new_rotated_gwcs = _rotated_gwcs(
        center_world_coord,
        rotation_angle,
        pixel_scales,
        cdelt_signs,
        refdata_shape=refdata_shape,
        image_shape=real_image_shape
    )

    # create a fake NDData (we use arange so data boundaries show up in Imviz
    # if it ever is accidentally exposed) with the rotated GWCS:
    placeholder_data = np.nan * np.ones(refdata_shape)

    ndd = NDData(
        data=placeholder_data,
        wcs=new_rotated_gwcs,
        meta={wcs_only_key: True, '_pixel_scales': pixel_scales}
    )
    return ndd


def _get_rotated_nddata_from_label(
    app, data_label, rotation_angle, refdata_shape=(10, 10),
    cdelt_signs=None, target_wcs_east_left=True, target_wcs_north_up=True
):
    """
    Create a synthetic NDData which stores GWCS that approximate
    the WCS in the coords attr of the Data object with label ``data_label``
    loaded into ``app``.

    This method is useful for rotating pre-loaded datasets when
    combined with ``app._change_reference_data(data_label)``.

    Parameters
    ----------
    app : `~jdaviz.Application`
        App instance containing ``data_label``.
    data_label : str
        Data label for the Data to rotate.
    rotation_angle : `~astropy.units.Quantity`
        Angle to rotate the image counter-clockwise from its
        original orientation.
    refdata_shape : tuple
        Shape of the reference data array.

    Returns
    -------
    ndd : `~astropy.nddata.NDData`
        Contains rotated WCS and meaningless data.

    Raises
    ------
    ValueError
        Data has no WCS.
    """
    data = app.data_collection[data_label]
    if data.coords is None:
        raise ValueError(f"{data_label} has no WCS for rotation.")

    # transform WCS relative to the first loaded data entry:
    wcs = data.coords
    degn, dege, flip = get_compass_info(data.coords, data.shape)[-3:]
    has_east_left = flip
    has_north_up = True  # assumed

    if isinstance(wcs, GWCS):
        lat_axis = wcs.world_axis_names.index("lat")
        lon_axis = wcs.world_axis_names.index("lon")
    else:  # FITS WCS
        lat_axis = wcs.wcs.lat
        lon_axis = wcs.wcs.lng

    if (
        not has_east_left and target_wcs_east_left and
        'imviz-compass' in [item['name'] for item in app.state.tray_items]
    ):
        # if an east/west flip is necessary, pass that along to the compass:
        compass_plugin = app.get_tray_item_from_name('imviz-compass')
        compass_plugin.flip_horizontal = not compass_plugin.flip_horizontal

    if cdelt_signs is None:
        cdelt_signs = [None, None]
        cdelt_signs[lon_axis] = (
            1 if ((has_east_left and target_wcs_east_left) or
                  (not has_east_left and not target_wcs_east_left)) else -1
        )
        cdelt_signs[lat_axis] = (
            1 if ((has_north_up and target_wcs_north_up) or
                  (not has_north_up and not target_wcs_north_up)) else -1
        )
    else:
        if has_east_left != target_wcs_east_left:
            cdelt_signs[lon_axis] = -1
        if has_north_up != target_wcs_north_up:
            cdelt_signs[lat_axis] = -1

    return _prepare_rotated_nddata(
        data.shape,
        data.coords,
        rotation_angle,
        refdata_shape,
        wcs_only_key=_wcs_only_label,
        data=data,
        cdelt_signs=cdelt_signs
    )


# This method comes from the jwst package:
# https://github.com/spacetelescope/jwst/blob/95467186aca9784ece9451b33d437d80d550a795/jwst/assign_wcs/util.py#L103
def compute_scale(wcs, fiducial, disp_axis, pscale_ratio=1):
    """Compute scaling transform.

    Parameters
    ----------
    wcs : `~gwcs.wcs.WCS`
        Reference WCS object from which to compute a scaling factor.

    fiducial : tuple
        Input fiducial of (RA, DEC) or (RA, DEC, Wavelength) used in calculating reference points.

    disp_axis : int
        Dispersion axis integer. Assumes the same convention as `wcsinfo.dispersion_direction`

    pscale_ratio : int
        Ratio of input to output pixel scale

    Returns
    -------
    scale : float
        Scaling factor for x and y or cross-dispersion direction.

    """
    spectral = 'SPECTRAL' in wcs.output_frame.axes_type

    if spectral and disp_axis is None:  # pragma: no cover
        raise ValueError('If input WCS is spectral, a disp_axis must be given')

    crpix = np.array(wcs.invert(*fiducial))

    delta = np.zeros_like(crpix)
    spatial_idx = np.where(np.array(wcs.output_frame.axes_type) == 'SPATIAL')[0]
    delta[spatial_idx[0]] = 1

    crpix_with_offsets = np.vstack((crpix, crpix + delta, crpix + np.roll(delta, 1))).T
    crval_with_offsets = wcs(*crpix_with_offsets, with_bounding_box=False)

    coords = SkyCoord(
        ra=crval_with_offsets[spatial_idx[0]],
        dec=crval_with_offsets[spatial_idx[1]],
        unit="deg"
    )
    xscale = np.abs(coords[0].separation(coords[1]).value)
    yscale = np.abs(coords[0].separation(coords[2]).value)

    if pscale_ratio is not None:
        xscale *= pscale_ratio
        yscale *= pscale_ratio

    if spectral:  # pragma: no cover
        # Assuming scale doesn't change with wavelength
        # Assuming disp_axis is consistent with DataModel.meta.wcsinfo.dispersion.direction
        return yscale if disp_axis == 1 else xscale

    return np.sqrt(xscale * yscale)



from pathlib import Path
import typing as t
import threading

from traitlets import Any, Dict, Instance, default
from ipywidgets import widget_serialization
from ipyvuetify import VuetifyTemplate

from jdaviz.core.style_widget import StyleWidget


T = t.TypeVar("T")
_style_paths: t.Dict[int, Path] = {}


# this is a temporary decorator until we can depend on solara, and solara has
# an equivalent implementation.
def _singleton(factory: [t.Callable[[], T]]) -> t.Callable[[], T]:
    def _get_unique_key():
        try:
            import solara.server.kernel_context as kernel_context

            try:
                kc = kernel_context.get_current_context()
                return kc.id
            except RuntimeError:
                pass  # not running in a solara virtual kernel
        except ModuleNotFoundError:
            return "not-in-solara"

    instances: t.Dict[str, T] = {}
    lock = threading.Lock()

    def wrapper(*args, **kwargs) -> T:
        key = _get_unique_key()
        if key not in instances:
            with lock:
                if key not in instances:
                    instances[key] = factory(*args, **kwargs)
        return instances[key]

    return wrapper


class StyleRegistry(VuetifyTemplate):
    style_widgets = Dict(default_value={}).tag(sync=True, **widget_serialization)

    @default("template")
    def _template(self):
        return """
        <template>
            <span>
                <jupyter-widget v-for="style_widget in style_widgets" :widget="style_widget">
                </jupyter-widget>
            </span>
        </template>
        """


def add(path):
    if hash(path) in _style_paths:
        return
    _style_paths[hash(path)] = path


class PopoutStyleWrapper(VuetifyTemplate):
    content = Any().tag(sync=True, **widget_serialization)
    style_registry_instance = Instance(
        StyleRegistry,
    ).tag(sync=True, **widget_serialization)

    @default("template")
    def _template(self):
        return """
        <template>
            <div>
                <jupyter-widget :widget="style_registry_instance"></jupyter-widget>
                <jupyter-widget :widget="content"></jupyter-widget>
            </div>
        </template>
        """

    @default("style_registry_instance")
    @_singleton
    def _default_style_registry_instance(self):
        return get_style_registry()


@_singleton
def get_style_registry():
    return StyleRegistry(
        style_widgets={key: StyleWidget(path) for key, path in _style_paths.items()}
    )




import pytest
import numpy as np
from astropy.io import fits
from jdaviz.core.data_formats import get_valid_format, identify_data
from jdaviz.core.config import list_configurations


def create_sdss(n_dim=2, **kwargs):
    """ create a fake SDSS single fiber spectral fits file """
    primary = fits.PrimaryHDU(header=fits.Header({'TELESCOP': "SDSS 2.5-M",
                                                  'FIBERID': 555,
                                                  'VERS2D': 'v5'}))
    ext = fits.BinTableHDU(name='DATA', header=fits.Header({'TTYPE3': "ivar"}))
    return fits.HDUList([primary, ext])


def create_manga(n_dim=None, **kwargs):
    """ create a fake SDSS MaNGA fits file """
    primary = fits.PrimaryHDU(header=fits.Header({'TELESCOP': "SDSS 2.5-M"}))
    ext = fits.ImageHDU(name='FLUX', header=fits.Header({'INSTRUME': "MaNGA"}))
    shape = (1,) if n_dim == 1 else (1, 2,) if n_dim == 2 else (1, 2, 3) if n_dim == 3 else (1,)
    ext.data = np.empty(shape)
    return fits.HDUList([primary, ext])


def _create_bintable(name, ver=1):
    return fits.BinTableHDU.from_columns(
                [fits.Column(name='target', format='20A', array=np.ones(3)),
                 fits.Column(name='V_mag', format='E', array=np.ones(3))],
                name=name, ver=ver
            )


def create_jwst(n_dim=None, multi=None, ext=None):
    """ create a fake JWST fits file """
    primary = fits.PrimaryHDU(header=fits.Header({'TELESCOP': "JWST"}))
    asdf = fits.ImageHDU(name='ASDF')
    sci = fits.ImageHDU(name='SCI')
    exts = [primary]
    if n_dim == 1:
        ex1d = _create_bintable(ext, ver=1)
        exts.append(ex1d)
        # create two more extensions
        if multi:
            for i in range(1, 3):
                ex1d = _create_bintable(ext, ver=i)
                exts.append(ex1d)
    shape = (1,) if n_dim == 1 else (1, 2,) if n_dim == 2 else (1, 2, 3) if n_dim == 3 else (1, )
    sci.data = np.empty(shape)
    if n_dim > 1:
        exts.append(sci)
    exts.append(asdf)
    return fits.HDUList(exts)


def create_generic(n_dim=1, **kwargs):
    """ create a generic fits file of dimension N """
    primary = fits.PrimaryHDU()
    ext = fits.ImageHDU(name='DATA')
    shape = (1,) if n_dim == 1 else (1, 2,) if n_dim == 2 else (1, 2, 3) if n_dim == 3 else (1,)
    ext.data = np.empty(shape)
    return fits.HDUList([primary, ext])


data = {'MaNGA cube': {'ndim': 3, 'fxn': create_manga},
        'MaNGA rss': {'ndim': 2, 'fxn': create_manga},
        'JWST s3d': {'ndim': 3, 'fxn': create_jwst},
        'JWST s2d': {'ndim': 2, 'fxn': create_jwst},
        'JWST x1d': {'ndim': 1, 'fxn': create_jwst},
        'JWST c1d': {'ndim': 1, 'fxn': create_jwst, 'ext': 'COMBINE1D'},
        'JWST x1d multi': {'ndim': 1, 'fxn': create_jwst, 'multi': True},
        'JWST c1d multi': {'ndim': 1, 'fxn': create_jwst, 'ext': 'COMBINE1D', 'multi': True},
        'SDSS-III/IV spec': {'ndim': 2, 'fxn': create_sdss},
        'generic 3d': {'ndim': 3, 'fxn': create_generic},
        'generic 1d': {'ndim': 1, 'fxn': create_generic}
        }


@pytest.fixture()
def create_fake_fits(tmp_path):
    """ fixture to create a fake FITS file in a temporary directory """
    def _create_fits(name):
        # generic fake file
        path = tmp_path / "fits"
        path.mkdir()
        filepath = path / (name.lower().replace('/', '_').replace(' ', '_') + '.fits')

        # generate fake fits.HDUList
        ndim = data[name]['ndim']
        fxn = data[name]['fxn']
        multi = data[name].get('multi', None)
        ext = data[name].get('ext', 'EXTRACT1D')
        hdulist = fxn(n_dim=ndim, ext=ext, multi=multi)
        hdulist.writeto(filepath)
        return str(filepath)

    return _create_fits


@pytest.mark.parametrize('name, expconf',
                         [('MaNGA cube', 'cubeviz'),
                          ('MaNGA rss', 'specviz'),
                          ('JWST s3d', 'cubeviz'),
                          ('JWST s2d', 'specviz2d'),
                          ('JWST x1d', 'specviz'),
                          ('JWST c1d', 'specviz'),
                          ('JWST x1d multi', 'specviz'),
                          ('JWST c1d multi', 'specviz'),
                          ('SDSS-III/IV spec', 'specviz'),
                          ('generic 3d', 'cubeviz'),
                          ('generic 1d', 'specviz')])
def test_get_valid_format(create_fake_fits, name, expconf):
    """ test correct file format and config is returned """
    filename = create_fake_fits(name)

    # for generic files that have no registered specutils format, valid_format will be blank
    name = name if 'generic' not in name else []

    valid_format, config = get_valid_format(filename)
    assert (valid_format, config) == (name, expconf)


def test_list_configurations():
    """ test correct configurations are listed """
    configs = list_configurations()
    assert set(configs).issubset({'default', 'cubeviz', 'specviz', 'mosviz',
                                  'imviz', 'specviz2d', 'rampviz'})


@pytest.mark.parametrize('name, expconf, expstat',
                         [('MaNGA cube', 'cubeviz', 'Success: Valid Format'),
                          ('generic 1d', 'specviz', 'Error: Cannot determine format of '
                                                    'the file to load.')])
def test_identify_data(create_fake_fits, name, expconf, expstat):
    """ test correct status messages are returned with identify_data """
    filename = create_fake_fits(name)
    name = name if 'generic' not in name else []

    try:
        valid_format, config = identify_data(filename)
        status = 'Success: Valid Format'
    except ValueError as e:
        valid_format, config = name, expconf
        status = f'Error: {e}'
    assert (valid_format, config) == (name, expconf)
    assert expstat in status


def test_identify_current_mismatch(create_fake_fits):
    """ test correct mismatch config status """
    filename = create_fake_fits('MaNGA cube')
    with pytest.raises(ValueError,
                       match='Mismatch between input file format and loaded configuration.'):
        identify_data(filename, current='specviz')



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

# NOTE: The configuration for the package, including the name, version, and
# other information are set in the setup.cfg file.

import os
import shutil
import sys

from setuptools import setup
from setuptools.command.develop import develop


# First provide helpful messages if contributors try and run legacy commands
# for tests or docs.

TEST_HELP = """
Note: running tests is no longer done using 'python setup.py test'. Instead
you will need to run:

    tox -e test

If you don't already have tox installed, you can install it with:

    pip install tox

If you only want to run part of the test suite, you can also use pytest
directly with::

    pip install -e .[test]
    pytest

For more information, see:

  http://docs.astropy.org/en/latest/development/testguide.html#running-tests
"""

if 'test' in sys.argv:
    print(TEST_HELP)
    sys.exit(1)

DOCS_HELP = """
Note: building the documentation is no longer done using
'python setup.py build_docs'. Instead you will need to
build the documentation with Sphinx directly using::

    pip install -e .[docs]
    cd docs
    make html

For more information, see:

  http://docs.astropy.org/en/latest/install.html#builddocs
"""

if 'build_docs' in sys.argv or 'build_sphinx' in sys.argv:
    print(DOCS_HELP)
    sys.exit(1)

VERSION_TEMPLATE = """
# Note that we need to fall back to the hard-coded version if either
# setuptools_scm can't be imported or setuptools_scm can't determine the
# version, so we catch the generic 'Exception'.
try:
    from setuptools_scm import get_version
    version = get_version(root='..', relative_to=__file__)
except Exception:
    version = '{version}'
""".lstrip()


# These are based on jupyter_core.paths
def jupyter_config_dir():
    """Get the Jupyter config directory for this platform and user.
    Returns JUPYTER_CONFIG_DIR if defined, else ~/.jupyter
    """
    import pathlib
    from tempfile import mkdtemp

    env = os.environ
    home_dir = pathlib.Path.home().as_posix()

    if env.get('JUPYTER_NO_CONFIG'):
        return mkdtemp(suffix='jupyter-clean-cfg')

    if env.get('JUPYTER_CONFIG_DIR'):
        return env['JUPYTER_CONFIG_DIR']

    return os.path.join(home_dir, '.jupyter')


def user_dir():
    homedir = os.path.expanduser('~')
    # Next line will make things work even when /home/ is a symlink to
    # /usr/home as it is on FreeBSD, for example
    homedir = os.path.realpath(homedir)
    if sys.platform == 'darwin':
        return os.path.join(homedir, 'Library', 'Jupyter')
    elif os.name == 'nt':
        appdata = os.environ.get('APPDATA', None)
        if appdata:
            return os.path.join(appdata, 'jupyter')
        else:
            return os.path.join(jupyter_config_dir(), 'data')
    else:
        # Linux, non-OS X Unix, AIX, etc.
        import pathlib
        env = os.environ
        home = pathlib.Path.home().as_posix()
        xdg = env.get("XDG_DATA_HOME", None)
        if not xdg:
            xdg = os.path.join(home, '.local', 'share')
        return os.path.join(xdg, 'jupyter')


class DevelopCmd(develop):
    prefix_targets = [
        (os.path.join("nbconvert", "templates"), 'jdaviz-default')
    ]

    def run(self):
        target_dir = os.path.join(sys.prefix, 'share', 'jupyter')
        if '--user' in sys.prefix:  # TODO: is there a better way to find out?
            target_dir = user_dir()
        target_dir = os.path.join(target_dir)

        for prefix_target, name in self.prefix_targets:
            source = os.path.join('share', 'jupyter', prefix_target, name)
            target = os.path.join(target_dir, prefix_target, name)
            target_subdir = os.path.dirname(target)
            if not os.path.exists(target_subdir):
                os.makedirs(target_subdir)
            abs_source = os.path.abspath(source)

            try:
                rel_source = os.path.relpath(abs_source, os.path.abspath(target_subdir))
            except Exception:
                # relpath does not work if source/target on different Windows disks.
                do_symlink = False
            else:
                do_symlink = True

            try:
                os.remove(target)
            except Exception:
                try:
                    shutil.rmtree(target)  # Maybe not a symlink
                except Exception:
                    pass

            # Cannot symlink without relpath or Windows admin priv in some OS versions.
            try:
                if do_symlink:
                    print(rel_source, '->', target)
                    os.symlink(rel_source, target)
                else:
                    raise OSError('just make copies')
            except Exception:
                print(abs_source, '->', target)
                shutil.copytree(abs_source, target)

        super(DevelopCmd, self).run()


# WARNING: all files generated during setup.py will not end up in the source
# distribution
data_files = []
# Add all the templates
for (dirpath, dirnames, filenames) in os.walk('share/jupyter/'):
    if filenames:
        data_files.append((dirpath, [os.path.join(dirpath, filename)
                                     for filename in filenames]))


setup(data_files=data_files, cmdclass={'develop': DevelopCmd},
      use_scm_version={'write_to': os.path.join('jdaviz', 'version.py'),
                       'write_to_template': VERSION_TEMPLATE})



import os
import warnings
from datetime import datetime, timezone

import numpy as np
from astropy import units as u
from astropy.modeling.fitting import TRFLSQFitter
from astropy.modeling import Parameter
from astropy.modeling.models import Gaussian1D
from astropy.time import Time
from glue.core.message import SubsetUpdateMessage
from ipywidgets import widget_serialization
from photutils.aperture import (ApertureStats, CircularAperture, EllipticalAperture,
                                RectangularAperture)
from traitlets import Any, Bool, Integer, List, Unicode, observe

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.custom_units_and_equivs import PIX2
from jdaviz.core.events import (GlobalDisplayUnitChanged, SnackbarMessage,
                                LinkUpdatedMessage, SliceValueUpdatedMessage)
from jdaviz.core.region_translators import regions2aperture, _get_region_from_spatial_subset
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, DatasetMultiSelectMixin,
                                        SubsetSelect, ApertureSubsetSelectMixin,
                                        TableMixin, PlotMixin, MultiselectMixin, with_spinner)
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               check_if_unit_is_per_solid_angle,
                                               flux_conversion_general,
                                               handle_squared_flux_unit_conversions)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.utils import PRIHDR_KEY

__all__ = ['SimpleAperturePhotometry']


@tray_registry('imviz-aper-phot-simple', label="Aperture Photometry")
class SimpleAperturePhotometry(PluginTemplateMixin, ApertureSubsetSelectMixin,
                               DatasetMultiSelectMixin, TableMixin, PlotMixin, MultiselectMixin):
    """
    The Aperture Photometry plugin performs aperture photometry for drawn regions.
    See the :ref:`Aperture Photometry Plugin Documentation <aper-phot-simple>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :meth:`~jdaviz.core.template_mixin.TableMixin.export_table`
    * :meth:`fitted_models`
    """
    template_file = __file__, "aper_phot_simple.vue"
    uses_active_status = Bool(True).tag(sync=True)

    aperture_area = Integer().tag(sync=True)
    background_items = List().tag(sync=True)
    background_selected = Unicode("").tag(sync=True)
    background_value = FloatHandleEmpty(0).tag(sync=True)
    pixel_area_multi_auto = Bool(True).tag(sync=True)
    pixel_area = FloatHandleEmpty(0).tag(sync=True)
    counts_factor = FloatHandleEmpty(0).tag(sync=True)
    flux_scaling_multi_auto = Bool(True).tag(sync=True)
    flux_scaling_warning = Unicode("").tag(sync=True)
    flux_scaling = FloatHandleEmpty(0).tag(sync=True)
    result_available = Bool(False).tag(sync=True)
    result_failed_msg = Unicode("").tag(sync=True)
    results = List().tag(sync=True)
    plot_types = List([]).tag(sync=True)
    current_plot_type = Unicode().tag(sync=True)
    plot_available = Bool(False).tag(sync=True)
    radial_plot = Any('').tag(sync=True, **widget_serialization)
    fit_radial_profile = Bool(False).tag(sync=True)
    fit_results = List().tag(sync=True)

    # Cubeviz only
    cube_slice = Unicode("").tag(sync=True)
    is_cube = Bool(False).tag(sync=True)

    # surface brightness display unit
    display_unit = Unicode("").tag(sync=True)

    # angle component of `display_unit`, to avoid repetition of seperating
    # it out from `display_unit`

    display_solid_angle_unit = Unicode("").tag(sync=True)

    # flux scaling display unit will always be flux, not sb. again its own
    # traitlet to avoid avoid repetition of seperating it out from `display_unit`
    flux_scaling_display_unit = Unicode("").tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Perform aperture photometry for drawn regions.'

        self.background = SubsetSelect(self,
                                       'background_items',
                                       'background_selected',
                                       dataset='dataset',
                                       default_text='Manual',
                                       manual_options=['Manual'],
                                       filters=['is_spatial', 'is_not_composite'])

        headers = ['xcenter', 'ycenter', 'sky_center',
                   'sum', 'sum_aper_area',
                   'aperture_sum_counts', 'aperture_sum_counts_err',
                   'aperture_sum_mag',
                   'min', 'max', 'mean', 'median', 'mode', 'std', 'mad_std', 'var',
                   'biweight_location', 'biweight_midvariance', 'fwhm',
                   'semimajor_sigma', 'semiminor_sigma', 'orientation', 'eccentricity',
                   'data_label', 'subset_label']
        self.table.headers_avail = headers
        self.table.headers_visible = headers

        self.plot_types = ["Curve of Growth", "Radial Profile", "Radial Profile (Raw)"]
        self.current_plot_type = self.plot_types[0]
        self._fitted_models = {}
        self._fitted_model_name = 'phot_radial_profile'

        # override default plot styling
        self.plot.figure.fig_margin = {'top': 60, 'bottom': 60, 'left': 65, 'right': 15}
        self.plot.viewer.axis_y.tick_format = '0.2e'
        self.plot.viewer.axis_y.label_offset = '55px'

        self.session.hub.subscribe(self, SubsetUpdateMessage, handler=self._on_subset_update)
        self.session.hub.subscribe(self, LinkUpdatedMessage, handler=self._on_link_update)

        # Custom dataset filters for Cubeviz
        if self.config == "cubeviz":
            def valid_cubeviz_datasets(data):

                comp = data.get_component(data.main_components[0])
                img_unit = u.Unit(comp.units) if comp.units else u.dimensionless_unscaled
                solid_angle_unit = check_if_unit_is_per_solid_angle(img_unit, return_unit=True)
                if solid_angle_unit is None:  # this is encountered sometimes ??
                    return

                # multiply out solid angle so we can check physical type of numerator
                img_unit *= solid_angle_unit

                acceptable_types = ['spectral flux density wav',
                                    'photon flux density wav',
                                    'spectral flux density',
                                    'photon flux density']

                return ((data.ndim in (2, 3)) and
                        (img_unit.physical_type in acceptable_types))

            self.dataset.add_filter(valid_cubeviz_datasets)
            self.session.hub.subscribe(self, SliceValueUpdatedMessage,
                                       handler=self._on_slice_changed)

            self.hub.subscribe(self, GlobalDisplayUnitChanged,
                               handler=self._on_display_units_changed)

    @property
    def user_api(self):
        # TODO: expose public API once finalized
        # expose=('multiselect', 'dataset', 'aperture',
        #                                   'background', 'background_value',
        #                                   'pixel_area', 'counts_factor', 'flux_scaling',
        #                                   'calculate_photometry',
        #                                   'unpack_batch_options', 'calculate_batch_photometry')

        return PluginUserApi(self, expose=('export_table', 'fitted_models'))

    @property
    def fitted_models(self):
        return self._fitted_models

    def _on_slice_changed(self, msg):
        if self.config != "cubeviz":
            return
        self.cube_slice = f"{msg.value:.3e} {msg.value_unit}"
        self._cube_wave = u.Quantity(msg.value, msg.value_unit)

    @observe("dataset_selected")
    def _on_dataset_selected_changed(self, event={}):
        if self.config != "cubeviz":
            return
        # self.dataset might not exist when app is setting itself up.
        if hasattr(self, "dataset"):
            if isinstance(self.dataset.selected_dc_item, list):
                datasets = self.dataset.selected_dc_item
            else:
                datasets = [self.dataset.selected_dc_item]

            self.is_cube = False
            for dataset in datasets:
                # This assumes all cubes, or no cubes. If we allow photometry on collapsed cubes
                # or images this will need to change.
                if dataset.ndim > 2:
                    self.is_cube = True
                    break

    def _on_display_units_changed(self, event={}):

        """
        Handle change of display units from Unit Conversion plugin (for now,
        cubeviz only). If new display units differ from input data units, input
        parameters for ap. phot. (i.e background, flux scaling) are converted
        to the new units. Photometry will remain in previous unit until
        'calculate' is pressed again.
        """

        if self.config == 'cubeviz':

            # get previously selected display units
            prev_display_unit = self.display_unit
            prev_flux_scale_unit = self.flux_scaling_display_unit

            # update display unit traitlets to new selection
            self._set_display_unit_of_selected_dataset()

            # convert the previous background and flux scaling values to new unit so
            # re-calculating photometry with the current selections will produce
            # the previous output with the new unit.
            if prev_display_unit != '':

                # convert background to new unit
                if self.background_value is not None:

                    # FIXME: Kinda hacky, better solution welcomed.
                    # Basically we want to forget about PIX2 and do normal conversion.
                    # Since traitlet does not keep unit, we do not need to reapply PIX2.
                    if "pix2" in prev_display_unit and "pix2" in self.display_unit:
                        prev_unit = u.Unit(prev_display_unit) * PIX2
                        new_unit = u.Unit(self.display_unit) * PIX2
                    else:
                        prev_unit = u.Unit(prev_display_unit)
                        new_unit = u.Unit(self.display_unit)

                    if self.multiselect:
                        if len(self.dataset.selected) == 1:
                            data = self.dataset.selected_dc_item[0]
                        else:
                            raise ValueError("cannot calculate background median in multiselect")
                    else:
                        data = self.dataset.selected_dc_item

                    if prev_unit != new_unit:

                        # NOTE: I don't think all of these equivalencies are necessary,
                        # but I'm keeping them since they were already here. Background
                        # should only be converted flux<>flux or sb<>sb so only a possible
                        # u.spectral_density would be needed. explore removing these as a follow up
                        pixar_sr = data.meta.get('_pixel_scale_factor', 1.0) if data else 1.0
                        equivs = all_flux_unit_conversion_equivs(pixar_sr,
                                                                 self._cube_wave) + u.spectral()

                        self.background_value = flux_conversion_general(self.background_value,
                                                                        prev_unit,
                                                                        new_unit,
                                                                        equivs,
                                                                        with_unit=False)

                # convert flux scaling to new unit
                if self.flux_scaling is not None:

                    prev_unit = u.Unit(prev_flux_scale_unit)
                    new_unit = u.Unit(self.flux_scaling_display_unit)
                    if prev_unit != new_unit:
                        equivs = u.spectral_density(self._cube_wave)
                        self.flux_scaling = flux_conversion_general(self.flux_scaling,
                                                                    prev_unit,
                                                                    new_unit,
                                                                    equivs,
                                                                    with_unit=False)

    def _set_display_unit_of_selected_dataset(self):

        """
        Set the display_unit and flux_scaling_display_unit traitlets,
        which depend on if the selected data set is flux or surface brightness,
        and the corresponding global display unit for either flux or
        surface brightness.
        """

        # all cubes are in sb so we can get display unit for plugin from SB display unit
        # this can be changed to listen specifically to changes in surface brightness
        # from UC plugin GlobalDisplayUnitChange message, but wiill require some refactoring
        disp_unit = self.app._get_display_unit('sb')

        # this check needs to be here because 'get_display_unit' will sometimes
        # return non surface brightness units or even None when the app is starting
        # up. this can be removed once that is fixed (see PR #3144)
        if disp_unit is None or not check_if_unit_is_per_solid_angle(disp_unit):
            self.display_unit = ''
            self.flux_scaling_display_unit = ''
            return

        self.display_unit = disp_unit

        # get angle componant of surface brightness
        # note: could add 'axis=angle' when cleaning this code up to avoid repeating this

        display_solid_angle_unit = check_if_unit_is_per_solid_angle(disp_unit, return_unit=True)
        if display_solid_angle_unit is not None:
            self.display_solid_angle_unit = display_solid_angle_unit.to_string()
        else:
            # there should always be a solid angle, but i think this is
            # encountered sometimes when initializing something..
            self.display_solid_angle_unit = ''

        # flux scaling will be applied when the solid angle componant is
        # multiplied out, so use 'flux' display unit
        fs_unit = self.app._get_display_unit('flux')
        self.flux_scaling_display_unit = fs_unit

        # if cube loaded is per-pixel-squared sb (i.e flux cube loaded)
        # pixel_area should be fixed to 1
        if self.display_solid_angle_unit == 'pix2':
            self.pixel_area = 1.0

    def _get_defaults_from_metadata(self, dataset=None):
        defaults = {}
        if dataset is None:
            meta = self.dataset.selected_dc_item.meta.copy()
        else:
            meta = self.dataset._get_dc_item(dataset).meta.copy()

        # Extract telescope specific unit conversion factors, if applicable.
        if PRIHDR_KEY in meta:
            meta.update(meta[PRIHDR_KEY])
            del meta[PRIHDR_KEY]
        if 'telescope' in meta:
            telescope = meta['telescope']
        else:
            telescope = meta.get('TELESCOP', '')
        if telescope == 'JWST':
            # Hardcode the flux conversion factor from MJy to ABmag
            mjy2abmag = 0.003631

            # if display unit is different, translate
            if (self.config == 'cubeviz') and (self.display_unit != ''):
                disp_unit = u.Unit(self.display_unit)
                mjy2abmag = flux_conversion_general(mjy2abmag,
                                                    u.MJy / u.sr,
                                                    disp_unit,
                                                    u.spectral_density(self._cube_wave),
                                                    with_unit=False)

            if 'photometry' in meta and 'pixelarea_arcsecsq' in meta['photometry']:
                defaults['pixel_area'] = meta['photometry']['pixelarea_arcsecsq']
                if 'bunit_data' in meta and meta['bunit_data'] == u.Unit("MJy/sr"):
                    defaults['flux_scaling'] = mjy2abmag
            elif 'PIXAR_A2' in meta:
                defaults['pixel_area'] = meta['PIXAR_A2']
                if 'BUNIT' in meta and meta['BUNIT'] == u.Unit("MJy/sr"):
                    defaults['flux_scaling'] = mjy2abmag

        elif telescope == 'HST':
            # TODO: Add more HST support, as needed.
            # HST pixel scales are from instrument handbooks.
            # This is really not used because HST data does not have sr in unit.
            # This is only for completeness.
            # For counts conversion, PHOTFLAM is used to convert "counts" to flux manually,
            # which is the opposite of JWST, so we just do not do it here.
            instrument = meta.get('INSTRUME', '').lower()
            detector = meta.get('DETECTOR', '').lower()
            if instrument == 'acs':
                if detector == 'wfc':
                    defaults['pixel_area'] = 0.05 * 0.05
                elif detector == 'hrc':  # pragma: no cover
                    defaults['pixel_area'] = 0.028 * 0.025
                elif detector == 'sbc':  # pragma: no cover
                    defaults['pixel_area'] = 0.034 * 0.03
            elif instrument == 'wfc3' and detector == 'uvis':  # pragma: no cover
                defaults['pixel_area'] = 0.04 * 0.04

        return defaults

    @observe('flux_scaling_multi_auto')
    def _multiselect_flux_scaling_warning(self, event={}):
        if not self.flux_scaling_multi_auto:
            self.flux_scaling_warning = ''
            return
        no_flux_scaling = [dataset for dataset in self.dataset.selected
                           if 'flux_scaling' not in self._get_defaults_from_metadata(dataset)]
        if len(no_flux_scaling):
            self.flux_scaling_warning = ('Could not determine flux scaling for '
                                         f'{", ".join(no_flux_scaling)}.  Those entries will '
                                         'default to zero.  Turn off auto-mode to provide '
                                         'flux-scaling manually.')
        else:
            self.flux_scaling_warning = ''

    @observe('flux_scaling')
    def _singleselect_flux_scaling_warning(self, event={}):
        if not self.multiselect:
            # disable warning once user changes value
            self.flux_scaling_warning = ''

    @observe('dataset_selected')
    def _dataset_selected_changed(self, event={}):
        if not hasattr(self, 'dataset'):
            # plugin not fully initialized
            return
        if self.dataset.selected_dc_item is None:
            return
        if self.multiselect:
            # defaults are applied within the loop if the auto-switches are enabled,
            # but we still need to update the flux-scaling warning
            self._multiselect_flux_scaling_warning()
            return

        try:
            defaults = self._get_defaults_from_metadata()
            self.counts_factor = 0
            self.pixel_area = defaults.get('pixel_area', 0)
            self.flux_scaling = defaults.get('flux_scaling', 0)
            if 'flux_scaling' in defaults:
                self.flux_scaling_warning = ''
            else:
                self.flux_scaling_warning = ('Could not determine flux scaling for '
                                             f'{self.dataset.selected}, defaulting to zero.')

        except Exception as e:
            self.hub.broadcast(SnackbarMessage(
                f"Failed to extract {self.dataset_selected}: {repr(e)}",
                color='error', sender=self))

        # get correct display unit for newly selected dataset
        if self.config == 'cubeviz':
            # sets display_unit and flux_scaling_display_unit traitlets
            self._set_display_unit_of_selected_dataset()

        # auto-populate background, if applicable.
        self._aperture_selected_changed()

    def _on_subset_update(self, msg):
        if not self.dataset_selected or not self.aperture_selected:
            return
        if self.multiselect:
            self._background_selected_changed()
            return

        sbst = msg.subset
        if sbst.label == self.aperture_selected and sbst.data.label == self.dataset_selected:
            self._aperture_selected_changed()
        elif sbst.label == self.background_selected and sbst.data.label == self.dataset_selected:
            self._background_selected_changed()

    def _on_link_update(self, msg):
        if not self.dataset_selected or not self.aperture_selected:
            return

        # Force background auto-calculation to update when linking has changed.
        self._aperture_selected_changed()

    @observe('aperture_selected', 'multiselect')
    def _aperture_selected_changed(self, event={}):
        if not self.dataset_selected or not self.aperture_selected:
            return
        if self.multiselect is not isinstance(self.aperture_selected, list):
            # then multiselect is in the process of changing but the traitlet for aperture_selected
            # has not been updated internally yet
            return
        if self.multiselect:
            self._background_selected_changed()
            return

        if self.config == 'cubeviz':
            self._set_display_unit_of_selected_dataset()

        # NOTE: aperture_selected can be triggered here before aperture_selected_validity is updated
        # so we'll still allow the snackbar to be raised as a second warning to the user and to
        # avoid acting on outdated information

        # NOTE: aperture area is only used to determine if a warning should be shown in the UI
        # and so does not need to be calculated within user API calls that don't act on traitlets
        try:
            # Sky subset does not have area. Not worth it to calculate just for a warning.
            if hasattr(self.aperture.selected_spatial_region, 'area'):
                self.aperture_area = int(np.ceil(self.aperture.selected_spatial_region.area))
            else:
                self.aperture_area = 0
        except Exception as e:
            self.hub.broadcast(SnackbarMessage(
                f"Failed to extract {self.aperture_selected}: {repr(e)}",
                color='error', sender=self))
        else:
            self._background_selected_changed()

    @property
    def _cubeviz_slice_ind(self):
        fv = self.app.get_viewer(self.app._jdaviz_helper._default_flux_viewer_reference_name)
        return fv.slice

    def _calc_background_median(self, reg, data=None):
        # Basically same way image stats are calculated in vue_do_aper_phot()
        # except here we only care about one stat for the background.
        if data is None:
            if self.multiselect:
                if len(self.dataset.selected) == 1:
                    data = self.dataset.selected_dc_item[0]
                else:
                    raise ValueError("cannot calculate background median in multiselect")
            else:
                data = self.dataset.selected_dc_item

        comp = data.get_component(data.main_components[0])

        if self.config == "cubeviz" and data.ndim > 2:
            comp_data = comp.data[:, :, self._cubeviz_slice_ind].T  # nx, ny --> ny, nx
            # Similar to coords_info logic.
            if '_orig_spec' in getattr(data, 'meta', {}):
                w = data.meta['_orig_spec'].wcs.celestial
            else:
                w = data.coords.celestial
        else:  # "imviz"
            comp_data = comp.data  # ny, nx
            w = data.coords

        if hasattr(reg, 'to_pixel'):
            reg = reg.to_pixel(w)
        aper_mask_stat = reg.to_mask(mode='center')
        img_stat = aper_mask_stat.get_values(comp_data, mask=None)

        # photutils/background/_utils.py --> nanmedian()
        bg_md = np.nanmedian(img_stat)  # Naturally in data unit

        # convert background median to display unit, if necessary (cubeviz only)
        if (self.config == 'cubeviz') and (self.display_unit != '') and comp.units:

            bg_md = flux_conversion_general(bg_md,
                                            u.Unit(comp.units),
                                            u.Unit(self.display_unit),
                                            u.spectral_density(self._cube_wave),
                                            with_unit=False)

        return bg_md

    @observe('background_selected')
    def _background_selected_changed(self, event={}):
        background_selected = event.get('new', self.background_selected)
        if background_selected == 'Manual':
            # we'll later access the user's self.background_value directly
            return

        if self.multiselect:
            # background_value will be recomputed within batch mode anyways and will
            # be replaced in the UI with a message
            self.background_value = -1
            return

        try:
            reg = _get_region_from_spatial_subset(self, self.background.selected_subset_state)
            self.background_value = self._calc_background_median(reg)
        except Exception as e:
            self.background_value = 0
            self.hub.broadcast(SnackbarMessage(
                f"Failed to extract {background_selected}: {repr(e)}", color='error', sender=self))

    @with_spinner()
    def calculate_photometry(self, dataset=None, aperture=None, background=None,
                             background_value=None, pixel_area=None, counts_factor=None,
                             flux_scaling=None, add_to_table=True, update_plots=True):
        """
        Calculate aperture photometry given the values set in the plugin or
        any overrides provided as arguments here (which will temporarily
        override plugin values for this calculation only).

        Note: Values set in the plugin in Cubeviz are in the selected display unit
        from the Unit conversion plugin. Overrides are, as the docstrings note,
        assumed to be in the units of the selected dataset.

        Parameters
        ----------
        dataset : str, optional
            Dataset to use for photometry.
        aperture : str, optional
            Subset to use as the aperture.
        background : str, optional
            Subset to use to calculate the background.
        background_value : float, optional
            Background to subtract, same unit as data.  Automatically computed if ``background``
            is set to a subset.
        pixel_area : float, optional
            Pixel area in arcsec squared, only used if data unit is a surface brightness unit.
        counts_factor : float, optional
            Factor to convert data unit to counts, in unit of flux/counts.
        flux_scaling : float, optional
            Same unit as data, used in -2.5 * log(flux / flux_scaling).
        add_to_table : bool, optional
        update_plots : bool, optional

        Returns
        -------
        table row, fit results
        """

        if self.multiselect and (dataset is None or aperture is None):  # pragma: no cover
            raise ValueError("for batch mode, use calculate_batch_photometry")

        if dataset is not None:
            if dataset not in self.dataset.choices:  # pragma: no cover
                raise ValueError(f"dataset must be one of {self.dataset.choices}")
            data = self.dataset._get_dc_item(dataset)
        else:
            # we can use the pre-cached value
            data = self.dataset.selected_dc_item

        if aperture is not None:
            if aperture not in self.aperture.choices:
                raise ValueError(f"aperture must be one of {self.aperture.choices}")

        if aperture is not None or dataset is not None:
            reg = self.aperture._get_spatial_region(subset=aperture if aperture is not None else self.aperture.selected,  # noqa
                                                    dataset=dataset if dataset is not None else self.dataset.selected)  # noqa
            # determine if a valid aperture (since selected_validity only applies to selected entry)
            _, _, validity = self.aperture._get_mark_coords_and_validate(selected=aperture)
            if not validity.get('is_aperture'):
                raise ValueError(f"Selected aperture {aperture} is not valid: {validity.get('aperture_message')}")  # noqa
        else:
            # use the pre-cached value
            if not self.aperture.selected_validity.get('is_aperture'):
                raise ValueError(f"Selected aperture is not valid: {self.aperture.selected_validity.get('aperture_message')}")  # noqa
            reg = self.aperture.selected_spatial_region

        # Reset last fitted model
        fit_model = None
        # TODO: remove _fitted_model_name cache?
        if self._fitted_model_name in self._fitted_models:
            del self._fitted_models[self._fitted_model_name]

        comp = data.get_component(data.main_components[0])
        if comp.units:
            img_unit = u.Unit(comp.units)
        else:
            img_unit = None

        if self.config == 'cubeviz':
            display_unit = u.Unit(self.display_unit)

        if background is not None and background not in self.background.choices:  # pragma: no cover
            raise ValueError(f"background must be one of {self.background.choices}")
        if background_value is not None:
            if ((background not in (None, 'Manual'))
                    or (background is None and self.background_selected != 'Manual')):
                raise ValueError("cannot provide background_value with background!='Manual'")
        elif (background == 'Manual'
                or (background is None and self.background.selected == 'Manual')):

            background_value = self.background_value

            # cubeviz: background_value set in plugin is in display units
            # convert temporarily to image units for calculations
            if (self.config == 'cubeviz') and (img_unit is not None) and display_unit != '':
                background_value = flux_conversion_general(background_value,
                                                           display_unit,
                                                           img_unit,
                                                           u.spectral_density(self._cube_wave),
                                                           with_unit=False)
        elif background is None and dataset is None:

            # use the previously-computed value in the plugin
            background_value = self.background_value

            # cubeviz: background_value set in plugin is in display units
            # convert temporarily to image units for calculations
            if (self.config == 'cubeviz') and (img_unit is not None) and display_unit != '':
                background_value = flux_conversion_general(background_value,
                                                           display_unit,
                                                           img_unit,
                                                           u.spectral_density(self._cube_wave),
                                                           with_unit=False)
        else:
            bg_reg = self.aperture._get_spatial_region(subset=background if background is not None else self.background.selected,  # noqa
                                                       dataset=dataset if dataset is not None else self.dataset.selected)  # noqa
            background_value = self._calc_background_median(bg_reg, data=data)

            # cubeviz: computed background median will be in display units,
            # convert temporarily back to image units for calculations
            if (self.config == 'cubeviz') and (img_unit is not None) and display_unit != '':
                background_value = flux_conversion_general(background_value,
                                                           display_unit,
                                                           img_unit,
                                                           u.spectral_density(self._cube_wave),
                                                           with_unit=False)
        try:
            bg = float(background_value)
        except ValueError:  # Clearer error message
            raise ValueError('Missing or invalid background value')

        if self.config == "cubeviz" and data.ndim > 2:
            comp_data = comp.data[:, :, self._cubeviz_slice_ind].T  # nx, ny --> ny, nx
            # Similar to coords_info logic.
            if '_orig_spec' in getattr(data, 'meta', {}):
                w = data.meta['_orig_spec'].wcs
            else:
                w = data.coords
        else:  # "imviz"
            comp_data = comp.data  # ny, nx
            w = data.coords

        if hasattr(reg, 'to_pixel'):
            sky_center = reg.center
            if self.config == "cubeviz" and data.ndim > 2:
                ycenter, xcenter = w.world_to_pixel(self._cube_wave, sky_center)[1]
            else:  # "imviz"
                xcenter, ycenter = w.world_to_pixel(sky_center)
        else:
            xcenter = reg.center.x
            ycenter = reg.center.y
            if data.coords is not None:
                if self.config == "cubeviz" and data.ndim > 2:
                    sky_center = w.pixel_to_world(self._cubeviz_slice_ind,
                                                  ycenter, xcenter)[1]
                else:  # "imviz"
                    sky_center = w.pixel_to_world(xcenter, ycenter)
            else:
                sky_center = None

        aperture = regions2aperture(reg)
        include_pixarea_fac = False
        include_counts_fac = False
        include_flux_scale = False
        if comp.units:

            # work for now in units of currently selected dataset (which may or
            # may not be the desired output units, depending on the display
            # units selected in the Unit Conversion plugin. background value
            # has already been converted to image units above, and flux scaling
            # will be converted from display unit > img_unit
            comp_data = comp_data << img_unit
            bg = bg * img_unit

            if check_if_unit_is_per_solid_angle(img_unit):  # if units are surface brightness
                try:
                    pixarea = float(pixel_area if pixel_area is not None else self.pixel_area)
                except ValueError:  # Clearer error message
                    raise ValueError('Missing or invalid pixel area')
                if not np.allclose(pixarea, 0):
                    include_pixarea_fac = True
            if img_unit != u.count:
                try:
                    ctfac = float(counts_factor if counts_factor is not None else self.counts_factor)  # noqa: E501
                except ValueError:  # Clearer error message
                    raise ValueError('Missing or invalid counts conversion factor')
                else:
                    if ctfac < 0:
                        raise ValueError('Counts conversion factor cannot be negative '
                                         f'but got {ctfac}.')
                if not np.allclose(ctfac, 0):
                    include_counts_fac = True

            # if cubeviz and flux_scaling is provided as override, it is in the data units
            # if set in the app, it is in the display units and needs to be converted
            # if provided as an override keyword arg, it is assumed to be in the
            # data units and does not need to be converted
            if ((self.config == 'cubeviz') and (flux_scaling is None) and
                    (self.flux_scaling is not None)):

                # convert flux_scaling from flux display unit to native flux unit
                flux_scaling = flux_conversion_general(
                    self.flux_scaling,
                    u.Unit(self.flux_scaling_display_unit),
                    img_unit * u.Unit(self.display_solid_angle_unit),
                    u.spectral_density(self._cube_wave),
                    with_unit=False)

            try:
                flux_scale = float(flux_scaling if flux_scaling is not None else self.flux_scaling)
            except ValueError:  # Clearer error message
                raise ValueError('Missing or invalid flux scaling')
            if not np.allclose(flux_scale, 0):
                include_flux_scale = True

            # from now, we will just need the image unit as a string for display
            img_unit = img_unit.to_string()

        else:
            img_unit = None

        phot_aperstats = ApertureStats(comp_data, aperture, wcs=data.coords, local_bkg=bg)
        phot_table = phot_aperstats.to_table(columns=(
            'id', 'sum', 'sum_aper_area',
            'min', 'max', 'mean', 'median', 'mode', 'std', 'mad_std', 'var',
            'biweight_location', 'biweight_midvariance', 'fwhm', 'semimajor_sigma',
            'semiminor_sigma', 'orientation', 'eccentricity'))  # Some cols excluded, add back as needed.  # noqa
        rawsum = phot_table['sum'][0]

        if include_pixarea_fac:
            # convert pixarea, which is in arcsec2/pix2 to the display solid angle unit / pix2

            if self.config == 'imviz':
                # can remove once unit conversion implemented in imviz and
                # display_solid_angle_unit traitlet is set, for now it will always be the data units
                display_solid_angle_unit = check_if_unit_is_per_solid_angle(comp.units,
                                                                            return_unit=True)

            elif self.config == 'cubeviz':
                display_solid_angle_unit = u.Unit(self.display_solid_angle_unit)

            # if angle unit is pix2, pixarea should be 1 pixel2 per pixel2
            if display_solid_angle_unit == PIX2:
                pixarea_fac = 1 * PIX2
            else:
                pixarea = pixarea * (u.arcsec * u.arcsec / PIX2)
                # NOTE: Sum already has npix value encoded, so we simply apply the npix unit here.
                # don't need to go though flux_conversion_general since these units
                # arent per-pixel and won't need a workaround.
                pixarea_fac = PIX2 * pixarea.to(display_solid_angle_unit / PIX2)

            phot_table['sum'] = [rawsum * pixarea_fac]
        else:
            pixarea_fac = None

        if include_counts_fac:
            ctfac = ctfac * (rawsum.unit / u.count)
            sum_ct = rawsum / ctfac
            sum_ct_err = np.sqrt(sum_ct.value) * sum_ct.unit
        else:
            ctfac = None
            sum_ct = None
            sum_ct_err = None

        if include_flux_scale:
            flux_scale = flux_scale * phot_table['sum'][0].unit
            sum_mag = -2.5 * np.log10(phot_table['sum'][0] / flux_scale) * u.mag
        else:
            flux_scale = None
            sum_mag = None

        # Extra info beyond photutils.
        phot_table.add_columns(
            [xcenter * u.pix, ycenter * u.pix, sky_center,
             bg, pixarea_fac, sum_ct, sum_ct_err, ctfac, sum_mag, flux_scale, data.label,
             reg.meta.get('label', ''), Time(datetime.now(tz=timezone.utc))],
            names=['xcenter', 'ycenter', 'sky_center', 'background', 'pixarea_tot',
                   'aperture_sum_counts', 'aperture_sum_counts_err', 'counts_fac',
                   'aperture_sum_mag', 'flux_scaling',
                   'data_label', 'subset_label', 'timestamp'],
            indexes=[1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 18, 18, 18])

        if self.config == "cubeviz":
            if data.ndim > 2:
                slice_val = self._cube_wave
            else:
                slice_val = u.Quantity(np.nan, self._cube_wave.unit)

            phot_table.add_column(slice_val, name="slice_wave", index=29)

            if comp.units:

                # convert units of output table to reflect display units
                # selected in Unit Conversion plugin
                display_unit = u.Unit(self.display_unit)

                # equivalencies for unit conversion, will never be flux<>sb
                # so only need spectral_density
                equivs = u.spectral_density(self._cube_wave)

                if display_unit != '':
                    if phot_table['background'].unit != display_unit:
                        bg_conv = flux_conversion_general(phot_table['background'].value,
                                                          phot_table['background'].unit,
                                                          display_unit,
                                                          equivs)
                        phot_table['background'] = bg_conv

                    phot_sum = phot_table['sum']
                    if include_pixarea_fac:
                        if phot_sum.unit != (display_unit * pixarea_fac).unit:
                            phot_table['sum'] = flux_conversion_general(phot_sum.value,
                                                                        phot_sum.unit,
                                                                        (display_unit * pixarea_fac).unit,  # noqa: E501
                                                                        equivs)

                    elif phot_sum.unit != display_unit:
                        phot_table['sum'] = flux_conversion_general(phot_sum.value,
                                                                    phot_sum.unit,
                                                                    display_unit,
                                                                    equivs)

                    for key in ['min', 'max', 'mean', 'median', 'mode', 'std',
                                'mad_std', 'biweight_location']:
                        if phot_table[key].unit != display_unit:
                            phot_table[key] = flux_conversion_general(phot_table[key].value,
                                                                      phot_table[key].unit,
                                                                      display_unit,
                                                                      equivs)

                    for key in ['var', 'biweight_midvariance']:
                        # these values will be in units of flux or surface brightness
                        # squared, so unit conversion is another special case if additional
                        # equivalencies are required
                        if phot_table[key].unit != display_unit**2:
                            conv = handle_squared_flux_unit_conversions(phot_table[key].value,
                                                                        phot_table[key].unit,
                                                                        display_unit**2,
                                                                        equivs)
                            phot_table[key] = conv

        if add_to_table:
            try:
                phot_table['id'][0] = self.table._qtable['id'].max() + 1
                self.table.add_item(phot_table)
            except Exception:  # Discard incompatible QTable
                self.table.clear_table()
                phot_table['id'][0] = 1
                self.table.add_item(phot_table)

            # User wants 'sum' as scientific notation.
            self.table._qtable['sum'].info.format = '.6e'

        # Plots.
        if update_plots:

            # for cubeviz unit conversion display units
            if self.display_unit != '':
                plot_display_unit = self.display_unit
            else:
                plot_display_unit = None

            if self.current_plot_type == "Curve of Growth":
                if self.config == "cubeviz" and data.ndim > 2:
                    self.plot.figure.title = f'Curve of growth from aperture center at {slice_val:.4e}'  # noqa: E501
                    eqv = u.spectral_density(self._cube_wave)
                else:
                    self.plot.figure.title = 'Curve of growth from aperture center'
                    eqv = []
                x_arr, sum_arr, x_label, y_label = _curve_of_growth(
                    comp_data, (xcenter, ycenter), aperture, phot_table['sum'][0],
                    wcs=data.coords, background=bg, pixarea_fac=pixarea_fac,
                    display_unit=plot_display_unit, equivalencies=eqv)
                self.plot._update_data('profile', x=x_arr, y=sum_arr, reset_lims=True)
                self.plot.update_style('profile', line_visible=True, color='gray', size=32)
                self.plot.update_style('fit', visible=False)
                self.plot.figure.axes[0].label = x_label
                self.plot.figure.axes[1].label = y_label

            else:  # Radial profile
                self.plot.figure.axes[0].label = 'pix'
                if plot_display_unit:
                    self.plot.figure.axes[1].label = plot_display_unit
                else:
                    self.plot.figure.axes[1].label = img_unit or 'Value'

                if self.current_plot_type == "Radial Profile":
                    if self.config == "cubeviz" and data.ndim > 2:
                        self.plot.figure.title = f'Radial profile from aperture center at {slice_val:.4e}'  # noqa: E501
                        eqv = u.spectral_density(self._cube_wave)
                    else:
                        self.plot.figure.title = 'Radial profile from aperture center'
                        eqv = []
                    x_data, y_data = _radial_profile(
                        phot_aperstats.data_cutout, phot_aperstats.bbox, (xcenter, ycenter),
                        raw=False, display_unit=plot_display_unit, image_unit=img_unit,
                        equivalencies=eqv)
                    self.plot._update_data('profile', x=x_data, y=y_data, reset_lims=True)
                    self.plot.update_style('profile', line_visible=True, color='gray', size=32)

                else:  # Radial Profile (Raw)
                    if self.config == "cubeviz" and data.ndim > 2:
                        self.plot.figure.title = f'Raw radial profile from aperture center at {slice_val:.4e}'  # noqa: E501
                    else:
                        self.plot.figure.title = 'Raw radial profile from aperture center'
                    x_data, y_data = _radial_profile(
                        phot_aperstats.data_cutout, phot_aperstats.bbox, (xcenter, ycenter),
                        raw=True, display_unit=plot_display_unit, image_unit=img_unit)

                    self.plot._update_data('profile', x=x_data, y=y_data, reset_lims=True)
                    self.plot.update_style('profile', line_visible=False, color='gray', size=10)

                # Fit Gaussian1D to radial profile data.
                if self.fit_radial_profile:
                    fitter = TRFLSQFitter()
                    y_max = np.nanmax(y_data)
                    x_mean = np.nanmean(x_data[np.where(y_data == y_max)])
                    std = 0.5 * (phot_table['semimajor_sigma'][0] +
                                 phot_table['semiminor_sigma'][0])
                    if isinstance(std, u.Quantity):
                        std = std.value
                    gs = Gaussian1D(amplitude=y_max, mean=x_mean, stddev=std,
                                    fixed={'amplitude': True},
                                    bounds={'amplitude': (y_max * 0.5, y_max)})
                    with warnings.catch_warnings(record=True) as warns:
                        fit_model = fitter(gs, x_data, y_data, filter_non_finite=True)
                    if len(warns) > 0:
                        msg = os.linesep.join([str(w.message) for w in warns])
                        self.hub.broadcast(SnackbarMessage(
                            f"Radial profile fitting: {msg}", color='warning', sender=self))
                    y_fit = fit_model(x_data)
                    self._fitted_models[self._fitted_model_name] = fit_model
                    self.plot._update_data('fit', x=x_data, y=y_fit, reset_lims=True)
                    self.plot.update_style('fit', color='magenta',
                                           markers_visible=False, line_visible=True)
                else:
                    self.plot.update_style('fit', visible=False)

        # Parse results for GUI.
        tmp = []
        for key in phot_table.colnames:
            if key in ('id', 'data_label', 'subset_label', 'background', 'pixarea_tot',
                       'counts_fac', 'aperture_sum_counts_err', 'flux_scaling', 'timestamp'):
                continue

            x = phot_table[key][0]

            if isinstance(x, u.Quantity):  # split up unit and value to put in different cols
                unit = x.unit.to_string()
                if unit == '':  # for eccentricity which is a quantity with an empty unit
                    unit = '-'
                x = x.value
            else:
                unit = '-'

            if (isinstance(x, (int, float)) and
                    key not in ('xcenter', 'ycenter', 'sky_center', 'sum_aper_area',
                                'aperture_sum_counts', 'aperture_sum_mag', 'slice_wave')):
                if x == 0:
                    tmp.append({'function': key, 'result': f'{x:.1f}', 'unit': unit})
                else:
                    tmp.append({'function': key, 'result': f'{x:.3e}', 'unit': unit})
            elif key == 'sky_center' and x is not None:
                tmp.append({'function': 'RA center', 'result': f'{x.ra.deg:.6f}', 'unit': 'deg'})
                tmp.append({'function': 'Dec center', 'result': f'{x.dec.deg:.6f}', 'unit': 'deg'})
            elif key in ('xcenter', 'ycenter', 'sum_aper_area'):
                tmp.append({'function': key, 'result': f'{x:.1f}', 'unit': unit})
            elif key == 'aperture_sum_counts' and x is not None:
                tmp.append({'function': key, 'result':
                            f'{x:.4e} ({phot_table["aperture_sum_counts_err"][0]:.4e})',
                            'unit': unit})
            elif key == 'aperture_sum_mag' and x is not None:
                tmp.append({'function': key, 'result': f'{x:.3f}', 'unit': unit})
            elif key == 'slice_wave':
                if data.ndim > 2:
                    tmp.append({'function': key, 'result': f'{slice_val.value:.4e}', 'unit': slice_val.unit.to_string()})  # noqa: E501
            else:
                tmp.append({'function': key, 'result': str(x), 'unit': unit})

        if update_plots:
            # Also display fit results
            fit_tmp = []
            if fit_model is not None and isinstance(fit_model, Gaussian1D):
                for param in ('mean', 'fwhm', 'amplitude'):
                    p_val = getattr(fit_model, param)
                    if isinstance(p_val, Parameter):
                        p_val = p_val.value
                    fit_tmp.append({'function': param, 'result': f'{p_val:.4e}'})

        self.results = tmp
        self.result_available = True

        if update_plots:
            self.fit_results = fit_tmp
            self.plot_available = True

        return phot_table, fit_model

    def vue_do_aper_phot(self, *args, **kwargs):
        if self.dataset_selected == '' or self.aperture_selected == '':
            self.hub.broadcast(SnackbarMessage(
                "No data for aperture photometry", color='error', sender=self))
            return

        try:
            if self.multiselect:
                # even though plots aren't show in the UI when in multiselect mode,
                # we'll create the last entry so if multiselect is disabled, the last
                # iteration will show and not result in confusing behavior
                self.calculate_batch_photometry(add_to_table=True, update_plots=True)
            else:
                self.calculate_photometry(add_to_table=True, update_plots=True)
        except Exception as e:  # pragma: no cover
            self.plot.clear_all_marks()
            msg = f"Aperture photometry failed: {repr(e)}"
            self.hub.broadcast(SnackbarMessage(msg, color='error', sender=self))
            self.result_failed_msg = msg
        else:
            self.result_failed_msg = ''

    def unpack_batch_options(self, **options):
        """
        Unpacks a dictionary of options for batch mode, including all combinations of any values
        passed as tuples or lists.  For example::

            unpack_batch_options(dataset=['image1', 'image2'],
                                 aperture=['Subset 1', 'Subset 2'],
                                 background=['Subset 3'],
                                 flux_scaling=3
                                 )

        would result in::

            [{'aperture': 'Subset 1',
              'dataset': 'image1',
              'background': 'Subset 3',
              'flux_scaling': 3},
             {'aperture': 'Subset 2',
              'dataset': 'image1',
              'background': 'Subset 3',
              'flux_scaling': 3},
             {'aperture': 'Subset 1',
              'dataset': 'image2',
              'background': 'Subset 3',
              'flux_scaling': 3},
             {'aperture': 'Subset 2',
              'dataset': 'image2',
              'background': 'Subset 3',
              'flux_scaling': 3}]

        Parameters
        ----------
        options : dict, optional
            Dictionary of values to override from the values set in the plugin/traitlets.  Each
            entry can either be a single value, or a list.  All combinations of those that contain
            a list will be exposed.  If not provided and the plugin is in multiselect mode
            (``multiselect = True``), then the current values set in the plugin will be used.

        Returns
        -------
        options : list
            List of all combinations of input parameters, which can then be used as input to
            `calculate_batch_photometry`.
        """
        if not isinstance(options, dict):
            raise TypeError("options must be a dictionary")
        if not options:
            if not self.multiselect:  # pragma: no cover
                raise ValueError("must either provide a dictionary or set plugin to multiselect mode")  # noqa
            options = {'dataset': self.dataset.selected, 'aperture': self.aperture.selected}

        # TODO: use self.user_api once API is made public
        user_api = self  # .user_api
        invalid_keys = [k for k in options.keys() if not hasattr(user_api, k)]
        if len(invalid_keys):
            raise ValueError(f"{invalid_keys} are not valid inputs for batch photometry")

        def _is_single(v):
            if isinstance(v, (list, tuple)):
                if len(v) == 1:
                    return True, v[0]
                return False, v
            return True, v

        single_values, mult_values = {}, {}
        for k, v in options.items():
            is_single, this_value = _is_single(v)
            if is_single:
                single_values[k] = this_value
            else:
                mult_values[k] = this_value

        def _unpack_dict_list(mult_values, single_values):
            if not len(mult_values):
                return [single_values]
            options_list = []
            # loop over the first item in mult_values
            # any remaining mult values will require recursion
            this_attr, this_values = list(mult_values.items())[0]
            remaining_mult_values = {k: v for j, (k, v) in enumerate(mult_values.items()) if j > 0}

            for this_value in this_values:
                if not len(remaining_mult_values):
                    options_list += [{this_attr: this_value, **single_values}]
                    continue
                options_list += _unpack_dict_list(remaining_mult_values,
                                                  {this_attr: this_value, **single_values})

            return options_list

        return _unpack_dict_list(mult_values, single_values)

    @with_spinner()
    def calculate_batch_photometry(self, options=[], add_to_table=True, update_plots=True,
                                   full_exceptions=False):
        """
        Run aperture photometry over a list of options.  Unprovided options will remain at their
        values defined in the plugin.

        To provide a list of values per-input, use `unpack_batch_options` to and pass that as input
        here.

        Parameters
        ----------
        options : list
            Each entry will result in one computation of aperture photometry and should be
            a dictionary of values to override from the values set in the plugin/traitlets.
        add_to_table : bool
            Whether to add results to the plugin table.
        update_plots : bool
            Whether to update the plugin plots for the last iteration.
        full_exceptions : bool, optional
            Whether to expose the full exception message for all failed iterations.
        """
        # input validation
        if not isinstance(options, list):
            raise TypeError("options must be a list of dictionaries")
        if not np.all([isinstance(option, dict) for option in options]):
            raise TypeError("options must be a list of dictionaries")
        if not len(options):
            if not self.multiselect:  # pragma: no cover
                raise ValueError("must either provide manual options or put the plugin in multiselect mode")  # noqa
            # unpack the batch options as provided in the app
            options = self.unpack_batch_options()

        failed_iters, exceptions = [], []
        for i, option in enumerate(options):
            # only update plots on the last iteration
            this_update_plots = i == len(options) and update_plots
            defaults = self._get_defaults_from_metadata(option.get('dataset',
                                                                   self.dataset.selected))
            if self.pixel_area_multi_auto:
                option.setdefault('pixel_area', defaults.get('pixel_area', 0))
            if self.flux_scaling_multi_auto:
                option.setdefault('flux_scaling', defaults.get('flux_scaling', 0))

            try:
                self.calculate_photometry(add_to_table=add_to_table,
                                          update_plots=this_update_plots,
                                          **option)
            except Exception as e:
                failed_iters.append(i)
                if full_exceptions:
                    exceptions.append(e)

        if len(failed_iters):
            err_msg = f"inputs {failed_iters} failed and were skipped."
            if full_exceptions:
                err_msg += f"  Exception messages: {exceptions}"
            else:
                err_msg += "  To see full exceptions, run individually or pass full_exceptions=True"  # noqa
            raise RuntimeError(err_msg)


# NOTE: These are hidden because the APIs are for internal use only
# but we need them as a separate functions for unit testing.

def _radial_profile(radial_cutout, reg_bb, centroid, raw=False,
                    image_unit=None, display_unit=None, equivalencies=[]):
    """Calculate radial profile.

    Parameters
    ----------
    radial_cutout : ndarray
        Cutout image from ``ApertureStats``.

    reg_bb : obj
        Bounding box from ``ApertureStats``.

    centroid : tuple of int
        ``ApertureStats`` centroid or desired center in ``(x, y)``.

    raw : bool
        If `True`, returns raw data points for scatter plot.
        Otherwise, use ``imexam`` algorithm for a clean plot.

    image_unit : str or None
        (For cubeviz only to deal with display unit conversion). Unit of input
        'radial cutout', used with `display_unit` to convert output to desired
        display unit.

    display_unit : str or None
        (For cubeviz only to deal with display unit conversion). Desired unit
        for output.

    equivalencies : list or None
        Optional, equivalencies for unit conversion to convert radial profile
        to display unit selected in the unit conversion plugin, if it differs
        from the native data unit.

    """
    reg_ogrid = np.ogrid[reg_bb.iymin:reg_bb.iymax, reg_bb.ixmin:reg_bb.ixmax]
    radial_dx = reg_ogrid[1] - centroid[0]
    radial_dy = reg_ogrid[0] - centroid[1]
    radial_r = np.hypot(radial_dx, radial_dy)

    # Sometimes the mask is smaller than radial_r
    if radial_cutout.shape != reg_bb.shape:
        radial_r = radial_r[:radial_cutout.shape[0], :radial_cutout.shape[1]]

    radial_r = radial_r[~radial_cutout.mask].ravel()  # pix
    radial_img = radial_cutout.compressed()  # data unit

    if raw:
        i_arr = np.argsort(radial_r)
        x_arr = radial_r[i_arr]
        y_arr = radial_img[i_arr]
    else:
        # This algorithm is from the imexam package,
        # see licenses/IMEXAM_LICENSE.txt for more details
        radial_r = np.rint(radial_r).astype(int)
        y_arr = np.bincount(radial_r, radial_img) / np.bincount(radial_r)
        x_arr = np.arange(y_arr.size)

    if display_unit is not None:
        if image_unit is None:
            raise ValueError('Must provide image_unit with display_unit.')

        # convert array from native data unit to display unit, if they differ
        if image_unit != display_unit:
            y_arr = flux_conversion_general(y_arr,
                                            u.Unit(image_unit),
                                            u.Unit(display_unit),
                                            equivalencies=equivalencies,
                                            with_unit=False)

    return x_arr, y_arr


def _curve_of_growth(data, centroid, aperture, final_sum, wcs=None, background=0,
                     n_datapoints=10, pixarea_fac=None, display_unit=None, equivalencies=[]):
    """Calculate curve of growth for aperture photometry.

    Parameters
    ----------
    data : ndarray or `~astropy.units.Quantity`
        Data for the calculation.

    centroid : tuple of int
        ``ApertureStats`` centroid or desired center in ``(x, y)``.

    aperture : obj
        ``photutils`` aperture to use, except its center will be
        changed to the given ``centroid``. This is because the aperture
        might be hand-drawn and a more accurate centroid has been
        recalculated separately.

    final_sum : float or `~astropy.units.Quantity`
        Aperture sum that is already calculated in the
        main plugin above.

    wcs : obj or `None`
        Supported WCS objects or `None`.

    background : float or `~astropy.units.Quantity`
        Background to subtract, if any. Unit must match ``data``.

    n_datapoints : int
        Number of data points in the curve.

    pixarea_fac : float or `None`
        For ``flux_unit/sr`` to ``flux_unit`` conversion.

    display_unit : str or None
        (For cubeviz only to deal with display unit conversion). Desired unit
        for output. If unit is a surface brightness, a Flux unit will be
        returned if pixarea_fac is provided.

    Returns
    -------
    x_arr : ndarray
        Data for X-axis of the curve.

    sum_arr : ndarray or `~astropy.units.Quantity`
        Data for Y-axis of the curve.

    x_label, y_label : str
        X- and Y-axis labels, respectively.

    Raises
    ------
    TypeError
        Unsupported aperture.

    """
    n_datapoints += 1  # n + 1

    # determined desired unit for output sum array and y label
    # cubeviz only to handle unit conversion display unit changes
    if display_unit is not None:
        sum_unit = u.Unit(display_unit)
    else:
        if isinstance(data, u.Quantity):
            sum_unit = data.unit
        else:
            sum_unit = None
    if sum_unit and pixarea_fac is not None:
        # multiply data unit by its solid angle to convert sum in sb to sum in flux
        sum_unit *= check_if_unit_is_per_solid_angle(sum_unit, return_unit=True)

    if hasattr(aperture, 'to_pixel'):
        aperture = aperture.to_pixel(wcs)

    if isinstance(aperture, CircularAperture):
        x_label = 'Radius (pix)'
        x_arr = np.linspace(0, aperture.r, num=n_datapoints)[1:]
        aper_list = [CircularAperture(centroid, cur_r) for cur_r in x_arr[:-1]]
    elif isinstance(aperture, EllipticalAperture):
        x_label = 'Semimajor axis (pix)'
        x_arr = np.linspace(0, aperture.a, num=n_datapoints)[1:]
        a_arr = x_arr[:-1]
        b_arr = aperture.b * a_arr / aperture.a
        aper_list = [EllipticalAperture(centroid, cur_a, cur_b, theta=aperture.theta)
                     for (cur_a, cur_b) in zip(a_arr, b_arr)]
    elif isinstance(aperture, RectangularAperture):
        x_label = 'Width (pix)'
        x_arr = np.linspace(0, aperture.w, num=n_datapoints)[1:]
        w_arr = x_arr[:-1]
        h_arr = aperture.h * w_arr / aperture.w
        aper_list = [RectangularAperture(centroid, cur_w, cur_h, theta=aperture.theta)
                     for (cur_w, cur_h) in zip(w_arr, h_arr)]
    else:
        raise TypeError(f'Unsupported aperture: {aperture}')

    sum_arr = [ApertureStats(data, cur_aper, wcs=wcs, local_bkg=background).sum
               for cur_aper in aper_list]
    if isinstance(sum_arr[0], u.Quantity):
        sum_arr = u.Quantity(sum_arr)
    else:
        sum_arr = np.array(sum_arr)
    if pixarea_fac is not None:
        sum_arr = sum_arr * pixarea_fac
    if isinstance(final_sum, u.Quantity):
        final_sum = flux_conversion_general(final_sum.value, final_sum.unit,
                                            sum_arr.unit, equivalencies)
    sum_arr = np.append(sum_arr, final_sum)

    if sum_unit is None:
        y_label = 'Value'
    else:
        y_label = sum_unit.to_string()
        # bqplot does not like Quantity
        sum_arr = flux_conversion_general(sum_arr.value, sum_arr.unit, sum_unit,
                                          equivalencies, with_unit=False)

    return x_arr, sum_arr, x_label, y_label



from traitlets import Int, Float

__all__ = ['HandleEmptyMixin', 'IntHandleEmpty', 'FloatHandleEmpty']


class HandleEmptyMixin:
    """Mixin to handle empty field."""
    def __init__(self, *args, **kwargs):
        self._empty_to_default = kwargs.pop('replace_with_default', False)
        super().__init__(*args, **kwargs)

    def validate(self, obj, value):
        if value is None or (isinstance(value, str) and not len(value)):
            if self._empty_to_default:
                # If the field is emptied, it will override with the default value.
                return self.default_value
            # The value will remain as the empty string or None, likely will need to either
            # couple this with form validation or handle the case where the value
            # is an empty string once retrieved.
            return value
        return super().validate(obj, value)


class IntHandleEmpty(HandleEmptyMixin, Int):
    """Mixin to handle empty integer field."""
    pass


class FloatHandleEmpty(HandleEmptyMixin, Float):
    """Mixin to handle empty floating point field."""
    def validate(self, obj, value):
        if value == '.':
            return value
        return super().validate(obj, value)



import numpy as np

from glue.core import Data
from jdaviz import Application

from regions import RectangleSkyRegion
from astropy.coordinates import Angle, SkyCoord
from astropy import units as u

header = {'S_REGION': 'POLYGON ICRS  5.029236065 4.992154276 5.029513148 '
          '4.992154276 5.029513148 4.992468585 5.029236065 4.992468585'}

skycoord = SkyCoord(5.02937461, 4.99231143,
                    unit=(u.Unit(u.deg),
                          u.Unit(u.deg)))

sky_region = RectangleSkyRegion(center=skycoord, width=Angle(0.0003143089999999251, u.deg),
                                height=Angle(0.00027603191980735836, u.deg))


def todo_fix_test_slit_overlay(spectral_cube_wcs):

    app = Application()
    dc = app.data_collection
    dc.append(Data(x=np.ones((3, 4, 5)), label='test', coords=spectral_cube_wcs))

    # so = SlitOverlay(app=app)
    #
    # so_sky_region = so.jwst_header_to_skyregion(header)
    #
    # assert np.allclose(sky_region.width.value, so_sky_region.width.value)
    # assert np.allclose(sky_region.height.value, so_sky_region.height.value)
    # assert sky_region.center.to_string() == so_sky_region.center.to_string()
